package iotpolicy;

import java_cup.runtime.ComplexSymbolFactory;
import java_cup.runtime.ScannerBuffer;
import java.io.*;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import iotpolicy.parser.Lexer;
import iotpolicy.parser.Parser;
import iotpolicy.tree.ParseNode;
import iotpolicy.tree.ParseNodeVector;
import iotpolicy.tree.ParseTreeHandler;
import iotpolicy.tree.Declaration;
import iotpolicy.tree.DeclarationHandler;
import iotpolicy.tree.CapabilityDecl;
import iotpolicy.tree.InterfaceDecl;
import iotpolicy.tree.RequiresDecl;
import iotpolicy.tree.EnumDecl;
import iotpolicy.tree.StructDecl;

import iotrmi.Java.IoTRMITypes;


/** Class IoTCompiler is the main interface/stub compiler for
 *  files generation. This class calls helper classes
 *  such as Parser, Lexer, InterfaceDecl, CapabilityDecl,
 *  RequiresDecl, ParseTreeHandler, etc.
 *
 * @author      Rahmadi Trimananda <rahmadi.trimananda @ uci.edu>
 * @version     1.0
 * @since       2016-09-22
 */
public class IoTCompiler {

	/**
	 * Class properties
	 */
	// Maps multiple interfaces to multiple objects of ParseTreeHandler
	private Map<String,ParseTreeHandler> mapIntfacePTH;
	private Map<String,DeclarationHandler> mapIntDeclHand;
	private Map<String,Map<String,Set<String>>> mapInt2NewInts;
	// Data structure to store our types (primitives and non-primitives) for compilation
	private Map<String,String> mapPrimitives;
	private Map<String,String> mapNonPrimitivesJava;
	private Map<String,String> mapNonPrimitivesCplus;
	// Other data structures
	private Map<String,Integer> mapIntfaceObjId;		// Maps interface name to object Id
	private Map<String,Integer> mapNewIntfaceObjId;		// Maps new interface name to its object Id (keep track of stubs)
	private PrintWriter pw;
	private String dir;
	private String subdir;

	/**
	 * Class constants
	 */
	private final static String OUTPUT_DIRECTORY = "output_files";

	private enum ParamCategory {

		PRIMITIVES,		// All the primitive types, e.g. byte, short, int, long, etc.
		NONPRIMITIVES,	// Non-primitive types, e.g. Set, Map, List, etc.
		USERDEFINED		// Non-supported type by default; assumed as driver classes
	}

	/**
	 * Class constructors
	 */
	public IoTCompiler() {

		mapIntfacePTH = new HashMap<String,ParseTreeHandler>();
		mapIntDeclHand = new HashMap<String,DeclarationHandler>();
		mapInt2NewInts = new HashMap<String,Map<String,Set<String>>>();
		mapIntfaceObjId = new HashMap<String,Integer>();
		mapNewIntfaceObjId = new HashMap<String,Integer>();
		mapPrimitives = new HashMap<String,String>();
			arraysToMap(mapPrimitives, IoTRMITypes.primitivesJava, IoTRMITypes.primitivesCplus);
		mapNonPrimitivesJava = new HashMap<String,String>();
			arraysToMap(mapNonPrimitivesJava, IoTRMITypes.nonPrimitivesJava, IoTRMITypes.nonPrimitiveJavaLibs);
		mapNonPrimitivesCplus = new HashMap<String,String>();
			arraysToMap(mapNonPrimitivesCplus, IoTRMITypes.nonPrimitivesJava, IoTRMITypes.nonPrimitivesCplus);
		pw = null;
		dir = OUTPUT_DIRECTORY;
		subdir = null;
	}


	/**
	 * setDataStructures() sets parse tree and other data structures based on policy files.
	 * <p>
	 * It also generates parse tree (ParseTreeHandler) and
	 * copies useful information from parse tree into
	 * InterfaceDecl, CapabilityDecl, and RequiresDecl 
	 * data structures.
	 * Additionally, the data structure handles are
	 * returned from tree-parsing for further process.
	 *
	 */
	public void setDataStructures(String origInt, ParseNode pnPol, ParseNode pnReq) {

		ParseTreeHandler ptHandler = new ParseTreeHandler(origInt, pnPol, pnReq);
		DeclarationHandler decHandler = new DeclarationHandler();
		// Process ParseNode and generate Declaration objects
		// Interface
		ptHandler.processInterfaceDecl();
		InterfaceDecl intDecl = ptHandler.getInterfaceDecl();
		decHandler.addInterfaceDecl(origInt, intDecl);
		// Capabilities
		ptHandler.processCapabilityDecl();
		CapabilityDecl capDecl = ptHandler.getCapabilityDecl();
		decHandler.addCapabilityDecl(origInt, capDecl);
		// Requires
		ptHandler.processRequiresDecl();
		RequiresDecl reqDecl = ptHandler.getRequiresDecl();
		decHandler.addRequiresDecl(origInt, reqDecl);
		// Enumeration
		ptHandler.processEnumDecl();
		EnumDecl enumDecl = ptHandler.getEnumDecl();
		decHandler.addEnumDecl(origInt, enumDecl);
		// Struct
		ptHandler.processStructDecl();
		StructDecl structDecl = ptHandler.getStructDecl();
		decHandler.addStructDecl(origInt, structDecl);

		mapIntfacePTH.put(origInt, ptHandler);
		mapIntDeclHand.put(origInt, decHandler);
		// Set object Id counter to 0 for each interface
		mapIntfaceObjId.put(origInt, new Integer(0));
	}


	/**
	 * getMethodsForIntface() reads for methods in the data structure
	 * <p>
	 * It is going to give list of methods for a certain interface
	 * 		based on the declaration of capabilities.
	 */
	public void getMethodsForIntface(String origInt) {

		ParseTreeHandler ptHandler = mapIntfacePTH.get(origInt);
		Map<String,Set<String>> mapNewIntMethods = new HashMap<String,Set<String>>();
		// Get set of new interfaces, e.g. CameraWithCaptureAndData
		// Generate this new interface with all the methods it needs
		//		from different capabilities it declares
		DeclarationHandler decHandler = mapIntDeclHand.get(origInt);
		RequiresDecl reqDecl = (RequiresDecl) decHandler.getRequiresDecl(origInt);
		Set<String> setIntfaces = reqDecl.getInterfaces();
		for (String strInt : setIntfaces) {

			// Initialize a set of methods
			Set<String> setMethods = new HashSet<String>();
			// Get list of capabilities, e.g. ImageCapture, VideoRecording, etc.
			List<String> listCapab = reqDecl.getCapabList(strInt);
			for (String strCap : listCapab) {

				// Get list of methods for each capability
				CapabilityDecl capDecl = (CapabilityDecl) decHandler.getCapabilityDecl(origInt);
				List<String> listCapabMeth = capDecl.getMethods(strCap);
				for (String strMeth : listCapabMeth) {

					// Add methods into setMethods
					// This is to also handle redundancies (say two capabilities
					//		share the same methods)
					setMethods.add(strMeth);
				}
			}
			// Add interface and methods information into map
			mapNewIntMethods.put(strInt, setMethods);
		}
		// Map the map of interface-methods to the original interface
		mapInt2NewInts.put(origInt, mapNewIntMethods);
	}


	/**
	 * HELPER: writeMethodJavaLocalInterface() writes the method of the interface
	 */
	private void writeMethodJavaLocalInterface(Collection<String> methods, InterfaceDecl intDecl) {

		for (String method : methods) {

			List<String> methParams = intDecl.getMethodParams(method);
			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			print("public " + intDecl.getMethodType(method) + " " +
				intDecl.getMethodId(method) + "(");
			for (int i = 0; i < methParams.size(); i++) {
				// Check for params with driver class types and exchange it 
				// 		with its remote interface
				String paramType = checkAndGetParamClass(methPrmTypes.get(i));
				print(paramType + " " + methParams.get(i));
				// Check if this is the last element (don't print a comma)
				if (i != methParams.size() - 1) {
					print(", ");
				}
			}
			println(");");
		}
	}


	/**
	 * HELPER: writeMethodJavaInterface() writes the method of the interface
	 */
	private void writeMethodJavaInterface(Collection<String> methods, InterfaceDecl intDecl) {

		for (String method : methods) {

			List<String> methParams = intDecl.getMethodParams(method);
			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			print("public " + intDecl.getMethodType(method) + " " +
				intDecl.getMethodId(method) + "(");
			for (int i = 0; i < methParams.size(); i++) {
				// Check for params with driver class types and exchange it 
				// 		with its remote interface
				String paramType = methPrmTypes.get(i);
				print(paramType + " " + methParams.get(i));
				// Check if this is the last element (don't print a comma)
				if (i != methParams.size() - 1) {
					print(", ");
				}
			}
			println(");");
		}
	}


	/**
	 * HELPER: writeEnumJava() writes the enumeration declaration
	 */
	private void writeEnumJava(EnumDecl enumDecl) {

		Set<String> enumTypes = enumDecl.getEnumDeclarations();
		// Iterate over enum declarations
		for (String enType : enumTypes) {

			println("public enum " + enType + " {");
			List<String> enumMembers = enumDecl.getMembers(enType);
			for (int i = 0; i < enumMembers.size(); i++) {

				String member = enumMembers.get(i);
				print(member);
				// Check if this is the last element (don't print a comma)
				if (i != enumMembers.size() - 1)
					println(",");
				else
					println("");
			}
			println("}\n");
		}
	}


	/**
	 * HELPER: writeStructJava() writes the struct declaration
	 */
	private void writeStructJava(StructDecl structDecl) {

		List<String> structTypes = structDecl.getStructTypes();
		// Iterate over enum declarations
		for (String stType : structTypes) {

			println("public class " + stType + " {");
			List<String> structMemberTypes = structDecl.getMemberTypes(stType);
			List<String> structMembers = structDecl.getMembers(stType);
			for (int i = 0; i < structMembers.size(); i++) {

				String memberType = structMemberTypes.get(i);
				String member = structMembers.get(i);
				println("public static " + memberType + " " + member + ";");
			}
			println("}\n");
		}
	}


	/**
	 * generateJavaLocalInterface() writes the local interface and provides type-checking.
	 * <p>
	 * It needs to rewrite and exchange USERDEFINED types in input parameters of stub
	 * and original interfaces, e.g. exchange Camera and CameraWithVideoAndRecording.
	 * The local interface has to be the input parameter for the stub and the stub 
	 * interface has to be the input parameter for the local class.
	 */
	public void generateJavaLocalInterfaces() throws IOException {

		// Create a new directory
		createDirectory(dir);
		for (String intface : mapIntfacePTH.keySet()) {
			// Open a new file to write into
			FileWriter fw = new FileWriter(dir + "/" + intface + ".java");
			pw = new PrintWriter(new BufferedWriter(fw));
			// Pass in set of methods and get import classes
			DeclarationHandler decHandler = mapIntDeclHand.get(intface);
			InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
			List<String> methods = intDecl.getMethods();
			Set<String> importClasses = getImportClasses(methods, intDecl);
			printImportStatements(importClasses);
			// Write interface header
			println("");
			println("public interface " + intface + " {");
			// Write enum if any...
			EnumDecl enumDecl = (EnumDecl) decHandler.getEnumDecl(intface);
			writeEnumJava(enumDecl);
			// Write struct if any...
			StructDecl structDecl = (StructDecl) decHandler.getStructDecl(intface);
			writeStructJava(structDecl);
			// Write methods
			writeMethodJavaLocalInterface(methods, intDecl);
			println("}");
			pw.close();
			System.out.println("IoTCompiler: Generated local interface " + intface + ".java...");
		}
	}


	/**
	 * generateJavaInterfaces() generate stub interfaces based on the methods list in Java
	 */
	public void generateJavaInterfaces() throws IOException {

		// Create a new directory
		String path = createDirectories(dir, subdir);
		for (String intface : mapIntfacePTH.keySet()) {

			Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
			for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {

				// Open a new file to write into
				String newIntface = intMeth.getKey();
				FileWriter fw = new FileWriter(path + "/" + newIntface + ".java");
				pw = new PrintWriter(new BufferedWriter(fw));
				DeclarationHandler decHandler = mapIntDeclHand.get(intface);
				InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
				// Pass in set of methods and get import classes
				Set<String> importClasses = getImportClasses(intMeth.getValue(), intDecl);
				printImportStatements(importClasses);
				// Write interface header
				println("");
				println("public interface " + newIntface + " {\n");
				// Write methods
				writeMethodJavaInterface(intMeth.getValue(), intDecl);
				println("}");
				pw.close();
				System.out.println("IoTCompiler: Generated interface " + newIntface + ".java...");
			}
		}
	}


	/**
	 * HELPER: writePropertiesJavaPermission() writes the permission in properties
	 */
	private void writePropertiesJavaPermission(String intface, InterfaceDecl intDecl) {

		Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
		for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {
			String newIntface = intMeth.getKey();
			int newObjectId = mapNewIntfaceObjId.get(newIntface);
			println("private final static int object" + newObjectId + "Id = " + 
				newObjectId + ";\t//" + newIntface);
			Set<String> methodIds = intMeth.getValue();
			print("private static Integer[] object" + newObjectId + "Permission = { ");
			int i = 0;
			for (String methodId : methodIds) {
				int methodNumId = intDecl.getMethodNumId(methodId);
				print(Integer.toString(methodNumId));
				// Check if this is the last element (don't print a comma)
				if (i != methodIds.size() - 1) {
					print(", ");
				}
				i++;
			}
			println(" };");
			println("private List<Integer> set" + newObjectId + "Allowed;");
		}
	}


	/**
	 * HELPER: writePropertiesJavaStub() writes the properties of the stub class
	 */
	private void writePropertiesJavaStub(String intface, String newIntface, boolean callbackExist, Set<String> callbackClasses) {

		println("private IoTRMICall rmiCall;");
		println("private String address;");
		println("private int[] ports;\n");
		// Get the object Id
		Integer objId = mapIntfaceObjId.get(intface);
		println("private final static int objectId = " + objId + ";");
		mapNewIntfaceObjId.put(newIntface, objId);
		mapIntfaceObjId.put(intface, objId++);
		if (callbackExist) {
		// We assume that each class only has one callback interface for now
			Iterator it = callbackClasses.iterator();
			String callbackType = (String) it.next();
			println("// Callback properties");
			println("private IoTRMIObject rmiObj;");
			println("List<" + callbackType + "> listCallbackObj;");
			println("private static int objIdCnt = 0;");
			// Generate permission stuff for callback stubs
			DeclarationHandler decHandler = mapIntDeclHand.get(callbackType);
			InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(callbackType);
			writePropertiesJavaPermission(callbackType, intDecl);
		}
		println("\n");
	}


	/**
	 * HELPER: writeConstructorJavaPermission() writes the permission in constructor
	 */
	private void writeConstructorJavaPermission(String intface) {

		Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
		for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {
			String newIntface = intMeth.getKey();
			int newObjectId = mapNewIntfaceObjId.get(newIntface);
			println("set" + newObjectId + "Allowed = Arrays.asList(object" + newObjectId +"Permission);");
		}
	}


	/**
	 * HELPER: writeConstructorJavaStub() writes the constructor of the stub class
	 */
	private void writeConstructorJavaStub(String intface, String newStubClass, boolean callbackExist, Set<String> callbackClasses) {

		println("public " + newStubClass + "(int _port, String _address, int _rev, int[] _ports) throws Exception {");
		println("address = _address;");
		println("ports = _ports;");
		println("rmiCall = new IoTRMICall(_port, _address, _rev);");
		if (callbackExist) {
			Iterator it = callbackClasses.iterator();
			String callbackType = (String) it.next();
			writeConstructorJavaPermission(intface);
			println("listCallbackObj = new ArrayList<" + callbackType + ">();");
			println("___initCallBack();");
		}
		println("}\n");
	}


	/**
	 * HELPER: writeJavaMethodCallbackPermission() writes permission checks in stub for callbacks
	 */
	private void writeJavaMethodCallbackPermission(String intface) {

		println("int methodId = IoTRMIObject.getMethodId(method);");
		// Get all the different stubs
		Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
		for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {
			String newIntface = intMeth.getKey();
			int newObjectId = mapNewIntfaceObjId.get(newIntface);
			println("if (!set" + newObjectId + "Allowed.contains(methodId)) {");
			println("throw new Error(\"Callback object for " + intface + " is not allowed to access method: \" + methodId);");
			println("}");
		}
	}


	/**
	 * HELPER: writeConstructorJavaStub() writes the constructor of the stub class
	 */
	private void writeInitCallbackJavaStub(String intface, InterfaceDecl intDecl) {

		println("public void ___initCallBack() {");
		// Generate main thread for callbacks
		println("Thread thread = new Thread() {");
		println("public void run() {");
		println("try {");
		println("rmiObj = new IoTRMIObject(ports[0]);");
		println("while (true) {");
		println("byte[] method = rmiObj.getMethodBytes();");
		writeJavaMethodCallbackPermission(intface);
		println("int objId = IoTRMIObject.getObjectId(method);");
		println(intface + "_CallbackSkeleton skel = (" + intface + "_CallbackSkeleton) listCallbackObj.get(objId);");
		println("if (skel != null) {");
		println("skel.invokeMethod(rmiObj);");
		println("} else {");
		println("throw new Error(\"" + intface + ": Object with Id \" + objId + \" not found!\");");
		println("}");
		println("}");
		println("} catch (Exception ex) {");
		println("ex.printStackTrace();");
		println("throw new Error(\"Error instantiating class " + intface + "_CallbackSkeleton!\");");
		println("}");
		println("}");
		println("};");
		println("thread.start();\n");
		// Generate info sending part
		String method = "___initCallBack()";
		println("int methodId = " + intDecl.getHelperMethodNumId(method) + ";");
		println("Class<?> retType = void.class;");
		println("Class<?>[] paramCls = new Class<?>[] { int.class, String.class, int.class };");
		println("Object[] paramObj = new Object[] { ports[0], address, 0 };");
		println("rmiCall.remoteCall(objectId, methodId, retType, null, paramCls, paramObj);");
		println("}\n");
	}


	/**
	 * HELPER: writeStdMethodBodyJavaStub() writes the standard method body in the stub class
	 */
	private void writeStdMethodBodyJavaStub(InterfaceDecl intDecl, List<String> methParams,
			List<String> methPrmTypes, String method) {

		println("int methodId = " + intDecl.getMethodNumId(method) + ";");
		String retType = intDecl.getMethodType(method);
		println("Class<?> retType = " + getSimpleType(retType) + ".class;");
		// Generate array of parameter types
		print("Class<?>[] paramCls = new Class<?>[] { ");
		for (int i = 0; i < methParams.size(); i++) {
			String paramType = checkAndGetArray(methPrmTypes.get(i), methParams.get(i));
			print(getSimpleType(paramType) + ".class");
			// Check if this is the last element (don't print a comma)
			if (i != methParams.size() - 1) {
				print(", ");
			}
		}
		println(" };");
		// Generate array of parameter objects
		print("Object[] paramObj = new Object[] { ");
		for (int i = 0; i < methParams.size(); i++) {
			print(getSimpleIdentifier(methParams.get(i)));
			// Check if this is the last element (don't print a comma)
			if (i != methParams.size() - 1) {
				print(", ");
			}
		}
		println(" };");
		// Check if this is "void"
		if (retType.equals("void")) {
			println("rmiCall.remoteCall(objectId, methodId, retType, null, paramCls, paramObj);");
		} else { // We do have a return value
		// Check if the return value NONPRIMITIVES
			if (getParamCategory(retType) == ParamCategory.NONPRIMITIVES) {
				String[] retGenValType = getTypeOfGeneric(retType);
				println("Class<?> retGenValType = " + retGenValType[0] + ".class;");
				println("Object retObj = rmiCall.remoteCall(objectId, methodId, retType, retGenValType, paramCls, paramObj);");
				println("return (" + retType + ")retObj;");
			} else {
				println("Object retObj = rmiCall.remoteCall(objectId, methodId, retType, null, paramCls, paramObj);");
				println("return (" + retType + ")retObj;");
			}
		}
	}


	/**
	 * HELPER: returnGenericCallbackType() returns the callback type
	 */
	private String returnGenericCallbackType(String paramType) {

		if (getParamCategory(paramType) == ParamCategory.NONPRIMITIVES)
			return getTypeOfGeneric(paramType)[0];
		else
			return paramType;
	}


	/**
	 * HELPER: checkCallbackType() checks the callback type
	 */
	private boolean checkCallbackType(String paramType, String callbackType) {

		String prmType = returnGenericCallbackType(paramType);
		return callbackType.equals(prmType);
	}


	/**
	 * HELPER: writeCallbackMethodBodyJavaStub() writes the callback method of the stub class
	 */
	private void writeCallbackMethodBodyJavaStub(InterfaceDecl intDecl, List<String> methParams,
			List<String> methPrmTypes, String method, String callbackType) {

		println("try {");
		// Check if this is single object, array, or list of objects
		for (int i = 0; i < methParams.size(); i++) {

			String paramType = methPrmTypes.get(i);
			if (checkCallbackType(paramType, callbackType)) { // Check if this has callback object
				String param = methParams.get(i);
				if (isArrayOrList(paramType, param)) {	// Generate loop
					println("for (" + paramType + " cb : " + getSimpleIdentifier(param) + ") {");
					println(callbackType + "_CallbackSkeleton skel = new " + callbackType + "_CallbackSkeleton(cb, objIdCnt++);");
				} else
					println(callbackType + "_CallbackSkeleton skel = new " + callbackType + "_CallbackSkeleton(" +
						getSimpleIdentifier(param) + ", objIdCnt++);");
				println("listCallbackObj.add(skel);");
				if (isArrayOrList(paramType, param))
					println("}");
			}
		}
		println("} catch (Exception ex) {");
		println("ex.printStackTrace();");
		println("throw new Error(\"Exception when generating skeleton objects!\");");
		println("}\n");
		println("int methodId = " + intDecl.getMethodNumId(method) + ";");
		String retType = intDecl.getMethodType(method);
		println("Class<?> retType = " + getSimpleType(retType) + ".class;");
		// Generate array of parameter types
		print("Class<?>[] paramCls = new Class<?>[] { ");
		for (int i = 0; i < methParams.size(); i++) {
			String paramType = methPrmTypes.get(i);
			if (checkCallbackType(paramType, callbackType)) { // Check if this has callback object
				print("int.class");
			} else { // Generate normal classes if it's not a callback object
				String prmType = checkAndGetArray(methPrmTypes.get(i), methParams.get(i));
				print(getSimpleType(prmType) + ".class");
			}
			if (i != methParams.size() - 1)	// Check if this is the last element
				print(", ");
		}
		println(" };");
		// Generate array of parameter objects
		print("Object[] paramObj = new Object[] { ");
		for (int i = 0; i < methParams.size(); i++) {
			String paramType = methPrmTypes.get(i);
			if (checkCallbackType(paramType, callbackType)) { // Check if this has callback object
				if (isArray(methPrmTypes.get(i), methParams.get(i)))
					print(getSimpleIdentifier(methParams.get(i)) + ".length");
				else if (isList(methPrmTypes.get(i), methParams.get(i)))
					print(getSimpleIdentifier(methParams.get(i)) + ".size()");
				else
					print("new Integer(1)");
			} else
				print(getSimpleIdentifier(methParams.get(i)));
			if (i != methParams.size() - 1)
				print(", ");
		}
		println(" };");
		// Check if this is "void"
		if (retType.equals("void")) {
			println("rmiCall.remoteCall(objectId, methodId, retType, null, paramCls, paramObj);");
		} else { // We do have a return value
		// Check if the return value NONPRIMITIVES
			if (getParamCategory(retType) == ParamCategory.NONPRIMITIVES) {
				String[] retGenValType = getTypeOfGeneric(retType);
				println("Class<?> retGenValType = " + retGenValType[0] + ".class;");
				println("Object retObj = rmiCall.remoteCall(objectId, methodId, retType, retGenValType, paramCls, paramObj);");
				println("return (" + retType + ")retObj;");
			} else {
				println("Object retObj = rmiCall.remoteCall(objectId, methodId, retType, null, paramCls, paramObj);");
				println("return (" + retType + ")retObj;");
			}
		}
	}


	/**
	 * HELPER: writeMethodJavaStub() writes the method of the stub class
	 */
	private void writeMethodJavaStub(Collection<String> methods, InterfaceDecl intDecl, Set<String> callbackClasses) {

		for (String method : methods) {

			List<String> methParams = intDecl.getMethodParams(method);
			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			print("public " + intDecl.getMethodType(method) + " " +
				intDecl.getMethodId(method) + "(");
			boolean isCallbackMethod = false;
			String callbackType = null;
			for (int i = 0; i < methParams.size(); i++) {

				String paramType = returnGenericCallbackType(methPrmTypes.get(i));
				// Check if this has callback object
				if (callbackClasses.contains(paramType)) {
					isCallbackMethod = true;
					callbackType = paramType;	
					// Even if there're 2 callback arguments, we expect them to be of the same interface
				}
				print(methPrmTypes.get(i) + " " + methParams.get(i));
				// Check if this is the last element (don't print a comma)
				if (i != methParams.size() - 1) {
					print(", ");
				}
			}
			println(") {");
			// Now, write the body of stub!
			if (isCallbackMethod)
				writeCallbackMethodBodyJavaStub(intDecl, methParams, methPrmTypes, method, callbackType);
			else
				writeStdMethodBodyJavaStub(intDecl, methParams, methPrmTypes, method);
			println("}\n");
			// Write the init callback helper method
			if (isCallbackMethod)
				writeInitCallbackJavaStub(callbackType, intDecl);
		}
	}


	/**
	 * generateJavaStubClasses() generate stubs based on the methods list in Java
	 */
	public void generateJavaStubClasses() throws IOException {

		// Create a new directory
		String path = createDirectories(dir, subdir);
		for (String intface : mapIntfacePTH.keySet()) {

			Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
			for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {

				// Open a new file to write into
				String newIntface = intMeth.getKey();
				String newStubClass = newIntface + "_Stub";
				FileWriter fw = new FileWriter(path + "/" + newStubClass + ".java");
				pw = new PrintWriter(new BufferedWriter(fw));
				DeclarationHandler decHandler = mapIntDeclHand.get(intface);
				InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
				// Pass in set of methods and get import classes
				Set<String> methods = intMeth.getValue();
				Set<String> importClasses = getImportClasses(methods, intDecl);
				List<String> stdImportClasses = getStandardJavaImportClasses();
				List<String> allImportClasses = getAllLibClasses(stdImportClasses, importClasses);
				printImportStatements(allImportClasses); println("");
				// Find out if there are callback objects
				Set<String> callbackClasses = getCallbackClasses(methods, intDecl);
				boolean callbackExist = !callbackClasses.isEmpty();
				// Write class header
				println("public class " + newStubClass + " implements " + newIntface + " {\n");
				// Write properties
				writePropertiesJavaStub(intface, newIntface, callbackExist, callbackClasses);
				// Write constructor
				writeConstructorJavaStub(intface, newStubClass, callbackExist, callbackClasses);
				// Write methods
				writeMethodJavaStub(intMeth.getValue(), intDecl, callbackClasses);
				println("}");
				pw.close();
				System.out.println("IoTCompiler: Generated stub class " + newStubClass + ".java...");
			}
		}
	}


	/**
	 * HELPER: writePropertiesJavaCallbackStub() writes the properties of the callback stub class
	 */
	private void writePropertiesJavaCallbackStub(String intface, String newIntface, boolean callbackExist, Set<String> callbackClasses) {

		println("private IoTRMICall rmiCall;");
		println("private String address;");
		println("private int[] ports;\n");
		// Get the object Id
		println("private static int objectId = 0;");
		if (callbackExist) {
		// We assume that each class only has one callback interface for now
			Iterator it = callbackClasses.iterator();
			String callbackType = (String) it.next();
			println("// Callback properties");
			println("private IoTRMIObject rmiObj;");
			println("List<" + callbackType + "> listCallbackObj;");
			println("private static int objIdCnt = 0;");
			// Generate permission stuff for callback stubs
			DeclarationHandler decHandler = mapIntDeclHand.get(callbackType);
			InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(callbackType);
			writePropertiesJavaPermission(callbackType, intDecl);
		}
		println("\n");
	}


	/**
	 * HELPER: writeConstructorJavaCallbackStub() writes the constructor of the callback stub class
	 */
	private void writeConstructorJavaCallbackStub(String intface, String newStubClass, boolean callbackExist, Set<String> callbackClasses) {

		// TODO: If we want callback in callback, then we need to add address and port initializations
		println("public " + newStubClass + "(IoTRMICall _rmiCall, int _objectId) throws Exception {");
		println("objectId = _objectId;");
		println("rmiCall = _rmiCall;");
		if (callbackExist) {
			Iterator it = callbackClasses.iterator();
			String callbackType = (String) it.next();
			writeConstructorJavaPermission(intface);
			println("listCallbackObj = new ArrayList<" + callbackType + ">();");
			println("___initCallBack();");
			println("// TODO: Add address and port initialization here if we want callback in callback!");
		}
		println("}\n");
	}


	/**
	 * generateJavaCallbackStubClasses() generate callback stubs based on the methods list in Java
	 * <p>
	 * Callback stubs gets the IoTRMICall objects from outside of the class as contructor input
	 * because all these stubs are populated by the class that takes in this object as a callback
	 * object. In such a class, we only use one socket, hence one IoTRMICall, for all callback objects.
	 */
	public void generateJavaCallbackStubClasses() throws IOException {

		// Create a new directory
		String path = createDirectories(dir, subdir);
		for (String intface : mapIntfacePTH.keySet()) {

			Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
			for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {

				// Open a new file to write into
				String newIntface = intMeth.getKey();
				String newStubClass = newIntface + "_CallbackStub";
				FileWriter fw = new FileWriter(path + "/" + newStubClass + ".java");
				pw = new PrintWriter(new BufferedWriter(fw));
				DeclarationHandler decHandler = mapIntDeclHand.get(intface);
				InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
				// Pass in set of methods and get import classes
				Set<String> methods = intMeth.getValue();
				Set<String> importClasses = getImportClasses(methods, intDecl);
				List<String> stdImportClasses = getStandardJavaImportClasses();
				List<String> allImportClasses = getAllLibClasses(stdImportClasses, importClasses);
				printImportStatements(allImportClasses); println("");
				// Find out if there are callback objects
				Set<String> callbackClasses = getCallbackClasses(methods, intDecl);
				boolean callbackExist = !callbackClasses.isEmpty();
				// Write class header
				println("public class " + newStubClass + " implements " + newIntface + " {\n");
				// Write properties
				writePropertiesJavaCallbackStub(intface, newIntface, callbackExist, callbackClasses);
				// Write constructor
				writeConstructorJavaCallbackStub(intface, newStubClass, callbackExist, callbackClasses);
				// Write methods
				// TODO: perhaps need to generate callback for callback
				writeMethodJavaStub(intMeth.getValue(), intDecl, callbackClasses);
				println("}");
				pw.close();
				System.out.println("IoTCompiler: Generated callback stub class " + newStubClass + ".java...");
			}
		}
	}


	/**
	 * HELPER: writePropertiesJavaSkeleton() writes the properties of the skeleton class
	 */
	private void writePropertiesJavaSkeleton(String intface, boolean callbackExist, InterfaceDecl intDecl) {

		println("private " + intface + " mainObj;");
		//println("private int ports;");
		println("private IoTRMIObject rmiObj;\n");
		// Callback
		if (callbackExist) {
			println("private static int objIdCnt = 0;");
			println("private IoTRMICall rmiCall;");
		}
		writePropertiesJavaPermission(intface, intDecl);
		println("\n");
	}


	/**
	 * HELPER: writeConstructorJavaSkeleton() writes the constructor of the skeleton class
	 */
	private void writeConstructorJavaSkeleton(String newSkelClass, String intface) {

		println("public " + newSkelClass + "(" + intface + " _mainObj, int _port) throws Exception {");
		println("mainObj = _mainObj;");
		println("rmiObj = new IoTRMIObject(_port);");
		// Generate permission control initialization
		writeConstructorJavaPermission(intface);
		println("___waitRequestInvokeMethod();");
		println("}\n");
	}


	/**
	 * HELPER: writeStdMethodBodyJavaSkeleton() writes the standard method body in the skeleton class
	 */
	private void writeStdMethodBodyJavaSkeleton(List<String> methParams, String methodId, String methodType) {

		if (methodType.equals("void"))
			print("mainObj." + methodId + "(");
		else
			print("return mainObj." + methodId + "(");
		for (int i = 0; i < methParams.size(); i++) {

			print(getSimpleIdentifier(methParams.get(i)));
			// Check if this is the last element (don't print a comma)
			if (i != methParams.size() - 1) {
				print(", ");
			}
		}
		println(");");
	}


	/**
	 * HELPER: writeInitCallbackJavaSkeleton() writes the init callback method for skeleton class
	 */
	private void writeInitCallbackJavaSkeleton(boolean callbackSkeleton) {

		// This is a callback skeleton generation
		if (callbackSkeleton)
			println("public void ___regCB(IoTRMIObject rmiObj) throws IOException {");
		else
			println("public void ___regCB() throws IOException {");
		println("Object[] paramObj = rmiObj.getMethodParams(new Class<?>[] { int.class, String.class, int.class },");
		println("\tnew Class<?>[] { null, null, null });");
		println("rmiCall = new IoTRMICall((int) paramObj[0], (String) paramObj[1], (int) paramObj[2]);");
		println("}\n");
	}


	/**
	 * HELPER: writeMethodJavaSkeleton() writes the method of the skeleton class
	 */
	private void writeMethodJavaSkeleton(Collection<String> methods, InterfaceDecl intDecl, Set<String> callbackClasses, 
			boolean callbackSkeleton) {

		for (String method : methods) {

			List<String> methParams = intDecl.getMethodParams(method);
			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			String methodId = intDecl.getMethodId(method);
			print("public " + intDecl.getMethodType(method) + " " + methodId + "(");
			boolean isCallbackMethod = false;
			String callbackType = null;
			for (int i = 0; i < methParams.size(); i++) {

				String origParamType = methPrmTypes.get(i);
				String paramType = checkAndGetParamClass(origParamType);
				if (callbackClasses.contains(origParamType)) { // Check if this has callback object
					isCallbackMethod = true;
					callbackType = origParamType;	
				}
				print(paramType + " " + methParams.get(i));
				// Check if this is the last element (don't print a comma)
				if (i != methParams.size() - 1) {
					print(", ");
				}
			}
			println(") {");
			// Now, write the body of skeleton!
			writeStdMethodBodyJavaSkeleton(methParams, methodId, intDecl.getMethodType(method));
			println("}\n");
			if (isCallbackMethod)
				writeInitCallbackJavaSkeleton(callbackSkeleton);
		}
	}


	/**
	 * HELPER: writeCallbackJavaStubGeneration() writes the callback stub generation part
	 */
	private Map<Integer,String> writeCallbackJavaStubGeneration(List<String> methParams, List<String> methPrmTypes, String callbackType) {

		Map<Integer,String> mapStubParam = new HashMap<Integer,String>();
		// Iterate over callback objects
		for (int i = 0; i < methParams.size(); i++) {
			String paramType = methPrmTypes.get(i);
			String param = methParams.get(i);
			//if (callbackType.equals(paramType)) {
			if (checkCallbackType(paramType, callbackType)) { // Check if this has callback object
				println("try {");
				String exchParamType = checkAndGetParamClass(paramType);
				// Print array if this is array or list if this is a list of callback objects
				if (isArray(paramType, param)) {
					println("int numStubs" + i + " = (int) paramObj[" + i + "];");
					println(exchParamType + "[] stub" + i + " = new " + exchParamType + "[numStubs" + i + "];");
				} else if (isList(paramType, param)) {
					println("int numStubs" + i + " = (int) paramObj[" + i + "];");
					println("List<" + exchParamType + "> stub" + i + " = new ArrayList<" + exchParamType + ">();");
				} else {
					println(exchParamType + " stub" + i + " = new " + exchParamType + "_CallbackStub(rmiCall, objIdCnt);");
					println("objIdCnt++;");
				}
			}
			// Generate a loop if needed
			if (checkCallbackType(paramType, callbackType)) { // Check if this has callback object
				String exchParamType = checkAndGetParamClass(paramType);
				if (isArray(paramType, param)) {
					println("for (int objId = 0; objId < numStubs" + i + "; objId++) {");
					println("stub" + i + "[objId] = new " + exchParamType + "_CallbackStub(rmiCall, objIdCnt);");
					println("objIdCnt++;");
					println("}");
				} else if (isList(paramType, param)) {
					println("for (int objId = 0; objId < numStubs" + i + "; objId++) {");
					println("stub" + i + ".add(new " + exchParamType + "_CallbackStub(rmiCall, objIdCnt));");
					println("objIdCnt++;");
					println("}");
				}
				mapStubParam.put(i, "stub" + i);	// List of all stub parameters
			}
		}
		return mapStubParam;
	}


	/**
	 * HELPER: writeStdMethodHelperBodyJavaSkeleton() writes the standard method body helper in the skeleton class
	 */
	private void writeStdMethodHelperBodyJavaSkeleton(InterfaceDecl intDecl, List<String> methParams,
			List<String> methPrmTypes, String method, Set<String> callbackClasses) {
		// Generate array of parameter objects
		boolean isCallbackMethod = false;
		String callbackType = null;
		print("Object[] paramObj = rmiObj.getMethodParams(new Class<?>[] { ");
		for (int i = 0; i < methParams.size(); i++) {

			String paramType = returnGenericCallbackType(methPrmTypes.get(i));
			if (callbackClasses.contains(paramType)) {
				isCallbackMethod = true;
				callbackType = paramType;
				print("int.class");
			} else {	// Generate normal classes if it's not a callback object
				String prmType = checkAndGetArray(methPrmTypes.get(i), methParams.get(i));
				print(getSimpleType(prmType) + ".class");
			}
			if (i != methParams.size() - 1)
				print(", ");
		}
		println(" }, ");
		// Generate generic class if it's a generic type.. null otherwise
		print("new Class<?>[] { ");
		for (int i = 0; i < methParams.size(); i++) {
			String prmType = methPrmTypes.get(i);
			if (getParamCategory(prmType) == ParamCategory.NONPRIMITIVES)
				print(getTypeOfGeneric(prmType)[0] + ".class");
			else
				print("null");
			if (i != methParams.size() - 1)
				print(", ");
		}
		println(" });");
		Map<Integer,String> mapStubParam = null;
		if (isCallbackMethod)
			mapStubParam = writeCallbackJavaStubGeneration(methParams, methPrmTypes, callbackType);
		// Check if this is "void"
		String retType = intDecl.getMethodType(method);
		if (retType.equals("void")) {
			print(intDecl.getMethodId(method) + "(");
		} else { // We do have a return value
			print("Object retObj = " + intDecl.getMethodId(method) + "(");
		}
		for (int i = 0; i < methParams.size(); i++) {

			if (isCallbackMethod) {
				print(mapStubParam.get(i));	// Get the callback parameter
			} else {
				String prmType = checkAndGetArray(methPrmTypes.get(i), methParams.get(i));
				print("(" + prmType + ") paramObj[" + i + "]");
			}
			if (i != methParams.size() - 1)
				print(", ");
		}
		println(");");
		if (!retType.equals("void"))
			println("rmiObj.sendReturnObj(retObj);");
		if (isCallbackMethod) {	// Catch exception if this is callback
			println("} catch(Exception ex) {");
			println("ex.printStackTrace();");
			println("throw new Error(\"Exception from callback object instantiation!\");");
			println("}");
		}
	}


	/**
	 * HELPER: writeMethodHelperJavaSkeleton() writes the method helper of the skeleton class
	 */
	private void writeMethodHelperJavaSkeleton(Collection<String> methods, InterfaceDecl intDecl, Set<String> callbackClasses) {

		// Use this set to handle two same methodIds
		Set<String> uniqueMethodIds = new HashSet<String>();
		for (String method : methods) {

			List<String> methParams = intDecl.getMethodParams(method);
			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			String methodId = intDecl.getMethodId(method);
			print("public void ___");
			String helperMethod = methodId;
			if (uniqueMethodIds.contains(methodId))
				helperMethod = helperMethod + intDecl.getMethodNumId(method);
			else
				uniqueMethodIds.add(methodId);
			// Check if this is "void"
			String retType = intDecl.getMethodType(method);
			if (retType.equals("void"))
				println(helperMethod + "() {");
			else
				println(helperMethod + "() throws IOException {");
			// Now, write the helper body of skeleton!
			writeStdMethodHelperBodyJavaSkeleton(intDecl, methParams, methPrmTypes, method, callbackClasses);
			println("}\n");
		}
	}


	/**
	 * HELPER: writeJavaMethodPermission() writes permission checks in skeleton
	 */
	private void writeJavaMethodPermission(String intface) {

		// Get all the different stubs
		Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
		for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {
			String newIntface = intMeth.getKey();
			int newObjectId = mapNewIntfaceObjId.get(newIntface);
			println("if (_objectId == object" + newObjectId + "Id) {");
			println("if (!set" + newObjectId + "Allowed.contains(methodId)) {");
			println("throw new Error(\"Object with object Id: \" + _objectId + \"  is not allowed to access method: \" + methodId);");
			println("}");
			println("else {");
			println("throw new Error(\"Object Id: \" + _objectId + \" not recognized!\");");
			println("}");
			println("}");
		}
	}


	/**
	 * HELPER: writeJavaWaitRequestInvokeMethod() writes the main loop of the skeleton class
	 */
	private void writeJavaWaitRequestInvokeMethod(Collection<String> methods, InterfaceDecl intDecl, boolean callbackExist, String intface) {

		// Use this set to handle two same methodIds
		Set<String> uniqueMethodIds = new HashSet<String>();
		println("private void ___waitRequestInvokeMethod() throws IOException {");
		// Write variables here if we have callbacks or enums or structs
		println("while (true) {");
		println("rmiObj.getMethodBytes();");
		println("int _objectId = rmiObj.getObjectId();");
		println("int methodId = rmiObj.getMethodId();");
		// Generate permission check
		writeJavaMethodPermission(intface);
		println("switch (methodId) {");
		// Print methods and method Ids
		for (String method : methods) {
			String methodId = intDecl.getMethodId(method);
			int methodNumId = intDecl.getMethodNumId(method);
			print("case " + methodNumId + ": ___");
			String helperMethod = methodId;
			if (uniqueMethodIds.contains(methodId))
				helperMethod = helperMethod + methodNumId;
			else
				uniqueMethodIds.add(methodId);
			println(helperMethod + "(); break;");
		}
		String method = "___initCallBack()";
		// Print case -9999 (callback handler) if callback exists
		if (callbackExist) {
			int methodId = intDecl.getHelperMethodNumId(method);
			println("case " + methodId + ": ___regCB(); break;");
		}
		println("default: ");
		println("throw new Error(\"Method Id \" + methodId + \" not recognized!\");");
		println("}");
		println("}");
		println("}\n");
	}


	/**
	 * generateJavaSkeletonClass() generate skeletons based on the methods list in Java
	 */
	public void generateJavaSkeletonClass() throws IOException {

		// Create a new directory
		String path = createDirectories(dir, subdir);
		for (String intface : mapIntfacePTH.keySet()) {
			// Open a new file to write into
			String newSkelClass = intface + "_Skeleton";
			FileWriter fw = new FileWriter(path + "/" + newSkelClass + ".java");
			pw = new PrintWriter(new BufferedWriter(fw));
			// Pass in set of methods and get import classes
			DeclarationHandler decHandler = mapIntDeclHand.get(intface);
			InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
			List<String> methods = intDecl.getMethods();
			Set<String> importClasses = getImportClasses(methods, intDecl);
			List<String> stdImportClasses = getStandardJavaImportClasses();
			List<String> allImportClasses = getAllLibClasses(stdImportClasses, importClasses);
			printImportStatements(allImportClasses);
			// Find out if there are callback objects
			Set<String> callbackClasses = getCallbackClasses(methods, intDecl);
			boolean callbackExist = !callbackClasses.isEmpty();
			// Write class header
			println("");
			println("public class " + newSkelClass  + " implements " + intface + " {\n");
			// Write properties
			writePropertiesJavaSkeleton(intface, callbackExist, intDecl);
			// Write constructor
			writeConstructorJavaSkeleton(newSkelClass, intface);
			// Write methods
			writeMethodJavaSkeleton(methods, intDecl, callbackClasses, false);
			// Write method helper
			writeMethodHelperJavaSkeleton(methods, intDecl, callbackClasses);
			// Write waitRequestInvokeMethod() - main loop
			writeJavaWaitRequestInvokeMethod(methods, intDecl, callbackExist, intface);
			println("}");
			pw.close();
			System.out.println("IoTCompiler: Generated skeleton class " + newSkelClass + ".java...");
		}
	}


	/**
	 * HELPER: writePropertiesJavaCallbackSkeleton() writes the properties of the callback skeleton class
	 */
	private void writePropertiesJavaCallbackSkeleton(String intface, boolean callbackExist) {

		println("private " + intface + " mainObj;");
		// For callback skeletons, this is its own object Id
		println("private static int objectId = 0;");
		// Callback
		if (callbackExist) {
			println("private static int objIdCnt = 0;");
			println("private IoTRMICall rmiCall;");
		}
		println("\n");
	}


	/**
	 * HELPER: writeConstructorJavaCallbackSkeleton() writes the constructor of the skeleton class
	 */
	private void writeConstructorJavaCallbackSkeleton(String newSkelClass, String intface) {

		println("public " + newSkelClass + "(" + intface + " _mainObj, int _objectId) throws Exception {");
		println("mainObj = _mainObj;");
		println("objectId = _objectId;");
		println("}\n");
	}


	/**
	 * HELPER: writeMethodHelperJavaCallbackSkeleton() writes the method helper of the callback skeleton class
	 */
	private void writeMethodHelperJavaCallbackSkeleton(Collection<String> methods, InterfaceDecl intDecl, Set<String> callbackClasses) {

		// Use this set to handle two same methodIds
		Set<String> uniqueMethodIds = new HashSet<String>();
		for (String method : methods) {

			List<String> methParams = intDecl.getMethodParams(method);
			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			String methodId = intDecl.getMethodId(method);
			print("public void ___");
			String helperMethod = methodId;
			if (uniqueMethodIds.contains(methodId))
				helperMethod = helperMethod + intDecl.getMethodNumId(method);
			else
				uniqueMethodIds.add(methodId);
			// Check if this is "void"
			String retType = intDecl.getMethodType(method);
			if (retType.equals("void"))
				println(helperMethod + "(IoTRMIObject rmiObj) {");
			else
				println(helperMethod + "(IoTRMIObject rmiObj) throws IOException {");
			// Now, write the helper body of skeleton!
			writeStdMethodHelperBodyJavaSkeleton(intDecl, methParams, methPrmTypes, method, callbackClasses);
			println("}\n");
		}
	}


	/**
	 * HELPER: writeJavaCallbackWaitRequestInvokeMethod() writes the main loop of the skeleton class
	 */
	private void writeJavaCallbackWaitRequestInvokeMethod(Collection<String> methods, InterfaceDecl intDecl, boolean callbackExist) {

		// Use this set to handle two same methodIds
		Set<String> uniqueMethodIds = new HashSet<String>();
		println("public void invokeMethod(IoTRMIObject rmiObj) throws IOException {");
		// Write variables here if we have callbacks or enums or structs
		println("int methodId = rmiObj.getMethodId();");
		// TODO: code the permission check here!
		println("switch (methodId) {");
		// Print methods and method Ids
		for (String method : methods) {
			String methodId = intDecl.getMethodId(method);
			int methodNumId = intDecl.getMethodNumId(method);
			print("case " + methodNumId + ": ___");
			String helperMethod = methodId;
			if (uniqueMethodIds.contains(methodId))
				helperMethod = helperMethod + methodNumId;
			else
				uniqueMethodIds.add(methodId);
			println(helperMethod + "(rmiObj); break;");
		}
		String method = "___initCallBack()";
		// Print case -9999 (callback handler) if callback exists
		if (callbackExist) {
			int methodId = intDecl.getHelperMethodNumId(method);
			println("case " + methodId + ": ___regCB(rmiObj); break;");
		}
		println("default: ");
		println("throw new Error(\"Method Id \" + methodId + \" not recognized!\");");
		println("}");
		println("}\n");
	}


	/**
	 * generateJavaCallbackSkeletonClass() generate callback skeletons based on the methods list in Java
	 */
	public void generateJavaCallbackSkeletonClass() throws IOException {

		// Create a new directory
		String path = createDirectories(dir, subdir);
		for (String intface : mapIntfacePTH.keySet()) {
			// Open a new file to write into
			String newSkelClass = intface + "_CallbackSkeleton";
			FileWriter fw = new FileWriter(path + "/" + newSkelClass + ".java");
			pw = new PrintWriter(new BufferedWriter(fw));
			// Pass in set of methods and get import classes
			DeclarationHandler decHandler = mapIntDeclHand.get(intface);
			InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
			List<String> methods = intDecl.getMethods();
			Set<String> importClasses = getImportClasses(methods, intDecl);
			List<String> stdImportClasses = getStandardJavaImportClasses();
			List<String> allImportClasses = getAllLibClasses(stdImportClasses, importClasses);
			printImportStatements(allImportClasses);
			// Find out if there are callback objects
			Set<String> callbackClasses = getCallbackClasses(methods, intDecl);
			boolean callbackExist = !callbackClasses.isEmpty();
			// Write class header
			println("");
			println("public class " + newSkelClass  + " implements " + intface + " {\n");
			// Write properties
			writePropertiesJavaCallbackSkeleton(intface, callbackExist);
			// Write constructor
			writeConstructorJavaCallbackSkeleton(newSkelClass, intface);
			// Write methods
			writeMethodJavaSkeleton(methods, intDecl, callbackClasses, true);
			// Write method helper
			writeMethodHelperJavaCallbackSkeleton(methods, intDecl, callbackClasses);
			// Write waitRequestInvokeMethod() - main loop
			writeJavaCallbackWaitRequestInvokeMethod(methods, intDecl, callbackExist);
			println("}");
			pw.close();
			System.out.println("IoTCompiler: Generated callback skeleton class " + newSkelClass + ".java...");
		}
	}


	/**
	 * HELPER: writeMethodCplusLocalInterface() writes the method of the interface
	 */
	private void writeMethodCplusLocalInterface(Collection<String> methods, InterfaceDecl intDecl) {

		for (String method : methods) {

			List<String> methParams = intDecl.getMethodParams(method);
			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			print("virtual " + checkAndGetCplusType(intDecl.getMethodType(method)) + " " +
				intDecl.getMethodId(method) + "(");
			for (int i = 0; i < methParams.size(); i++) {
				// Check for params with driver class types and exchange it 
				// 		with its remote interface
				String paramType = checkAndGetParamClass(methPrmTypes.get(i));
				paramType = checkAndGetCplusType(paramType);
				// Check for arrays - translate into vector in C++
				String paramComplete = checkAndGetCplusArray(paramType, methParams.get(i));
				print(paramComplete);
				// Check if this is the last element (don't print a comma)
				if (i != methParams.size() - 1) {
					print(", ");
				}
			}
			println(") = 0;");
		}
	}


	/**
	 * HELPER: writeMethodCplusInterface() writes the method of the interface
	 */
	private void writeMethodCplusInterface(Collection<String> methods, InterfaceDecl intDecl) {

		for (String method : methods) {

			List<String> methParams = intDecl.getMethodParams(method);
			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			print("virtual " + checkAndGetCplusType(intDecl.getMethodType(method)) + " " +
				intDecl.getMethodId(method) + "(");
			for (int i = 0; i < methParams.size(); i++) {
				// Check for params with driver class types and exchange it 
				// 		with its remote interface
				String paramType = methPrmTypes.get(i);
				paramType = checkAndGetCplusType(paramType);
				// Check for arrays - translate into vector in C++
				String paramComplete = checkAndGetCplusArray(paramType, methParams.get(i));
				print(paramComplete);
				// Check if this is the last element (don't print a comma)
				if (i != methParams.size() - 1) {
					print(", ");
				}
			}
			println(") = 0;");
		}
	}


	/**
	 * HELPER: writeEnumCplus() writes the enumeration declaration
	 */
	private void writeEnumCplus(EnumDecl enumDecl) {

		Set<String> enumTypes = enumDecl.getEnumDeclarations();
		// Iterate over enum declarations
		for (String enType : enumTypes) {

			println("enum " + enType + " {");
			List<String> enumMembers = enumDecl.getMembers(enType);
			for (int i = 0; i < enumMembers.size(); i++) {

				String member = enumMembers.get(i);
				print(member);
				// Check if this is the last element (don't print a comma)
				if (i != enumMembers.size() - 1)
					println(",");
				else
					println("");
			}
			println("};\n");
		}
	}


	/**
	 * HELPER: writeStructCplus() writes the struct declaration
	 */
	private void writeStructCplus(StructDecl structDecl) {

		List<String> structTypes = structDecl.getStructTypes();
		// Iterate over enum declarations
		for (String stType : structTypes) {

			println("struct " + stType + " {");
			List<String> structMemberTypes = structDecl.getMemberTypes(stType);
			List<String> structMembers = structDecl.getMembers(stType);
			for (int i = 0; i < structMembers.size(); i++) {

				String memberType = structMemberTypes.get(i);
				String member = structMembers.get(i);
				String structTypeC = checkAndGetCplusType(memberType);
				String structComplete = checkAndGetCplusArray(structTypeC, member);
				println(structComplete + ";");
			}
			println("};\n");
		}
	}


	/**
	 * generateCplusLocalInterfaces() writes the local interfaces and provides type-checking.
	 * <p>
	 * It needs to rewrite and exchange USERDEFINED types in input parameters of stub
	 * and original interfaces, e.g. exchange Camera and CameraWithVideoAndRecording.
	 * The local interface has to be the input parameter for the stub and the stub 
	 * interface has to be the input parameter for the local class.
	 */
	public void generateCplusLocalInterfaces() throws IOException {

		// Create a new directory
		createDirectory(dir);
		for (String intface : mapIntfacePTH.keySet()) {
			// Open a new file to write into
			FileWriter fw = new FileWriter(dir + "/" + intface + ".hpp");
			pw = new PrintWriter(new BufferedWriter(fw));
			// Write file headers
			println("#ifndef _" + intface.toUpperCase() + "_HPP__");
			println("#define _" + intface.toUpperCase() + "_HPP__");
			println("#include <iostream>");
			// Pass in set of methods and get include classes
			DeclarationHandler decHandler = mapIntDeclHand.get(intface);
			InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
			List<String> methods = intDecl.getMethods();
			Set<String> includeClasses = getIncludeClasses(methods, intDecl, true);
			printIncludeStatements(includeClasses); println("");
			println("using namespace std;\n");
			// Write enum if any...
			EnumDecl enumDecl = (EnumDecl) decHandler.getEnumDecl(intface);
			writeEnumCplus(enumDecl);
			// Write struct if any...
			StructDecl structDecl = (StructDecl) decHandler.getStructDecl(intface);
			writeStructCplus(structDecl);
			println("class " + intface); println("{");
			println("public:");
			// Write methods
			writeMethodCplusLocalInterface(methods, intDecl);
			println("};");
			println("#endif");
			pw.close();
			System.out.println("IoTCompiler: Generated local interface " + intface + ".hpp...");
		}
	}


	/**
	 * generateCPlusInterfaces() generate stub interfaces based on the methods list in C++
	 * <p>
	 * For C++ we use virtual classe as interface
	 */
	public void generateCPlusInterfaces() throws IOException {

		// Create a new directory
		String path = createDirectories(dir, subdir);
		for (String intface : mapIntfacePTH.keySet()) {

			Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
			for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {

				// Open a new file to write into
				String newIntface = intMeth.getKey();
				FileWriter fw = new FileWriter(path + "/" + newIntface + ".hpp");
				pw = new PrintWriter(new BufferedWriter(fw));
				DeclarationHandler decHandler = mapIntDeclHand.get(intface);
				InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
				// Write file headers
				println("#ifndef _" + newIntface.toUpperCase() + "_HPP__");
				println("#define _" + newIntface.toUpperCase() + "_HPP__");
				println("#include <iostream>");
				// Pass in set of methods and get import classes
				Set<String> includeClasses = getIncludeClasses(intMeth.getValue(), intDecl, false);
				List<String> stdIncludeClasses = getStandardCplusIncludeClasses();
				List<String> allIncludeClasses = getAllLibClasses(stdIncludeClasses, includeClasses);
				printIncludeStatements(allIncludeClasses); println("");			
				println("using namespace std;\n");
				println("class " + newIntface);
				println("{");
				println("public:");
				// Write methods
				writeMethodCplusInterface(intMeth.getValue(), intDecl);
				println("};");
				println("#endif");
				pw.close();
				System.out.println("IoTCompiler: Generated interface " + newIntface + ".hpp...");
			}
		}
	}


	/**
	 * HELPER: writeMethodCplusStub() writes the method of the stub
	 */
	private void writeMethodCplusStub(Collection<String> methods, InterfaceDecl intDecl, Set<String> callbackClasses) {

		for (String method : methods) {

			List<String> methParams = intDecl.getMethodParams(method);
			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			print(checkAndGetCplusType(intDecl.getMethodType(method)) + " " +
				intDecl.getMethodId(method) + "(");
			boolean isCallbackMethod = false;
			String callbackType = null;
			for (int i = 0; i < methParams.size(); i++) {

				String paramType = methPrmTypes.get(i);
				// Check if this has callback object
				if (callbackClasses.contains(paramType)) {
					isCallbackMethod = true;
					callbackType = paramType;	
					// Even if there're 2 callback arguments, we expect them to be of the same interface
				}
				String methPrmType = checkAndGetCplusType(methPrmTypes.get(i));
				String methParamComplete = checkAndGetCplusArray(methPrmType, methParams.get(i));
				print(methParamComplete);
				// Check if this is the last element (don't print a comma)
				if (i != methParams.size() - 1) {
					print(", ");
				}
			}
			println(") { ");
			if (isCallbackMethod)
				writeCallbackMethodBodyCplusStub(intDecl, methParams, methPrmTypes, method, callbackType);
			else
				writeStdMethodBodyCplusStub(intDecl, methParams, methPrmTypes, method);
			println("}\n");
			// Write the init callback helper method
			if (isCallbackMethod) {
				writeInitCallbackCplusStub(callbackType, intDecl);
				writeInitCallbackSendInfoCplusStub(intDecl);
			}
		}
	}


	/**
	 * HELPER: writeCallbackMethodBodyCplusStub() writes the callback method of the stub class
	 */
	private void writeCallbackMethodBodyCplusStub(InterfaceDecl intDecl, List<String> methParams,
			List<String> methPrmTypes, String method, String callbackType) {

		// Check if this is single object, array, or list of objects
		boolean isArrayOrList = false;
		String callbackParam = null;
		for (int i = 0; i < methParams.size(); i++) {

			String paramType = methPrmTypes.get(i);
			if (checkCallbackType(paramType, callbackType)) { // Check if this has callback object
				String param = methParams.get(i);
				if (isArrayOrList(paramType, param)) {	// Generate loop
					println("for (" + paramType + "* cb : " + getSimpleIdentifier(param) + ") {");
					println(callbackType + "_CallbackSkeleton* skel = new " + callbackType + "_CallbackSkeleton(cb, objIdCnt++);");
					isArrayOrList = true;
					callbackParam = getSimpleIdentifier(param);
				} else
					println(callbackType + "_CallbackSkeleton* skel = new " + callbackType + "_CallbackSkeleton(" +
						getSimpleIdentifier(param) + ", objIdCnt++);");
				println("vecCallbackObj.push_back(skel);");
				if (isArrayOrList(paramType, param))
					println("}");
			}
		}
		println("int numParam = " + methParams.size() + ";");
		println("int methodId = " + intDecl.getMethodNumId(method) + ";");
		String retType = intDecl.getMethodType(method);
		String retTypeC = checkAndGetCplusType(retType);
		println("string retType = \"" + checkAndGetCplusArrayType(retTypeC) + "\";");
		// Generate array of parameter types
		print("string paramCls[] = { ");
		for (int i = 0; i < methParams.size(); i++) {
			String paramType = methPrmTypes.get(i);
			if (checkCallbackType(paramType, callbackType)) { // Check if this has callback object
				print("\"int\"");
			} else { // Generate normal classes if it's not a callback object
				String paramTypeC = checkAndGetCplusType(methPrmTypes.get(i));
				String prmType = checkAndGetCplusArrayType(paramTypeC, methParams.get(i));
				print("\"" + prmType + "\"");
			}
			if (i != methParams.size() - 1)	// Check if this is the last element
				print(", ");
		}
		println(" };");
		print("int ___paramCB = ");
		if (isArrayOrList)
			println(callbackParam + ".size();");
		else
			println("1;");
		// Generate array of parameter objects
		print("void* paramObj[] = { ");
		for (int i = 0; i < methParams.size(); i++) {
			String paramType = methPrmTypes.get(i);
			if (checkCallbackType(paramType, callbackType)) { // Check if this has callback object
				print("&___paramCB");
			} else
				print(getSimpleIdentifier(methParams.get(i)));
			if (i != methParams.size() - 1)
				print(", ");
		}
		println(" };");
		// Check if this is "void"
		if (retType.equals("void")) {
			println("void* retObj = NULL;");
			println("rmiCall->remoteCall(objectId, methodId, retType, paramCls, paramObj, numParam, retObj);");
		} else { // We do have a return value
			if (getParamCategory(retType) == ParamCategory.NONPRIMITIVES)
				println(checkAndGetCplusType(retType) + " retVal;");
			else
				println(checkAndGetCplusType(retType) + " retVal = " + generateCplusInitializer(retType) + ";");
			println("void* retObj = &retVal;");
			println("rmiCall->remoteCall(objectId, methodId, retType, paramCls, paramObj, numParam, retObj);");
			println("return retVal;");
		}
	}


	/**
	 * HELPER: writeStdMethodBodyCplusStub() writes the standard method body in the stub class
	 */
	private void writeStdMethodBodyCplusStub(InterfaceDecl intDecl, List<String> methParams,
			List<String> methPrmTypes, String method) {

		println("int numParam = " + methParams.size() + ";");
		println("int methodId = " + intDecl.getMethodNumId(method) + ";");
		String retType = intDecl.getMethodType(method);
		String retTypeC = checkAndGetCplusType(retType);
		println("string retType = \"" + checkAndGetCplusArrayType(retTypeC) + "\";");
		// Generate array of parameter types
		print("string paramCls[] = { ");
		for (int i = 0; i < methParams.size(); i++) {
			String paramTypeC = checkAndGetCplusType(methPrmTypes.get(i));
			String paramType = checkAndGetCplusArrayType(paramTypeC, methParams.get(i));
			print("\"" + paramType + "\"");
			// Check if this is the last element (don't print a comma)
			if (i != methParams.size() - 1) {
				print(", ");
			}
		}
		println(" };");
		// Generate array of parameter objects
		print("void* paramObj[] = { ");
		for (int i = 0; i < methParams.size(); i++) {
			print("&" + getSimpleIdentifier(methParams.get(i)));
			// Check if this is the last element (don't print a comma)
			if (i != methParams.size() - 1) {
				print(", ");
			}
		}
		println(" };");
		// Check if this is "void"
		if (retType.equals("void")) {
			println("void* retObj = NULL;");
			println("rmiCall->remoteCall(objectId, methodId, retType, paramCls, paramObj, numParam, retObj);");
		} else { // We do have a return value
			if (getParamCategory(retType) == ParamCategory.NONPRIMITIVES)
				println(checkAndGetCplusType(retType) + " retVal;");
			else
				println(checkAndGetCplusType(retType) + " retVal = " + generateCplusInitializer(retType) + ";");
			println("void* retObj = &retVal;");
			println("rmiCall->remoteCall(objectId, methodId, retType, paramCls, paramObj, numParam, retObj);");
			println("return retVal;");
		}
	}


	/**
	 * HELPER: writePropertiesCplusStub() writes the properties of the stub class
	 */
	private void writePropertiesCplusPermission(String intface) {

		Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
		for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {
			String newIntface = intMeth.getKey();
			int newObjectId = mapNewIntfaceObjId.get(newIntface);
			println("const static int object" + newObjectId + "Id = " + newObjectId + ";");
			println("const static set<int> set" + newObjectId + "Allowed;");
		}
	}	

	/**
	 * HELPER: writePropertiesCplusStub() writes the properties of the stub class
	 */
	private void writePropertiesCplusStub(String intface, String newIntface, boolean callbackExist, Set<String> callbackClasses) {

		println("IoTRMICall *rmiCall;");
		//println("IoTRMIObject\t\t\t*rmiObj;");
		println("string address;");
		println("vector<int> ports;\n");
		// Get the object Id
		Integer objId = mapIntfaceObjId.get(intface);
		println("const static int objectId = " + objId + ";");
		mapNewIntfaceObjId.put(newIntface, objId);
		mapIntfaceObjId.put(intface, objId++);
		if (callbackExist) {
		// We assume that each class only has one callback interface for now
			Iterator it = callbackClasses.iterator();
			String callbackType = (String) it.next();
			println("// Callback properties");
			println("IoTRMIObject *rmiObj;");
			println("vector<" + callbackType + "*> vecCallbackObj;");
			println("static int objIdCnt;");
			// Generate permission stuff for callback stubs
			writePropertiesCplusPermission(callbackType);
		}
		println("\n");
	}


	/**
	 * HELPER: writeConstructorCplusStub() writes the constructor of the stub class
	 */
	private void writeConstructorCplusStub(String newStubClass, boolean callbackExist, Set<String> callbackClasses) {

		println(newStubClass + 
			"(int _port, const char* _address, int _rev, bool* _bResult, vector<int> _ports) {");
		println("address = _address;");
		println("ports = _ports;");
		println("rmiCall = new IoTRMICall(_port, _address, _rev, _bResult);");
		if (callbackExist) {
			println("objIdCnt = 0;");
			Iterator it = callbackClasses.iterator();
			String callbackType = (String) it.next();
			println("thread th1 (&" + newStubClass + "::___initCallBack, this);");
			println("th1.detach();");
			println("___regCB();");
		}
		println("}\n");
	}


	/**
	 * HELPER: writeDeconstructorCplusStub() writes the deconstructor of the stub class
	 */
	private void writeDeconstructorCplusStub(String newStubClass, boolean callbackExist, Set<String> callbackClasses) {

		println("~" + newStubClass + "() {");
		println("if (rmiCall != NULL) {");
		println("delete rmiCall;");
		println("rmiCall = NULL;");
		println("}");
		if (callbackExist) {
		// We assume that each class only has one callback interface for now
			println("if (rmiObj != NULL) {");
			println("delete rmiObj;");
			println("rmiObj = NULL;");
			println("}");
			Iterator it = callbackClasses.iterator();
			String callbackType = (String) it.next();
			println("for(" + callbackType + "* cb : vecCallbackObj) {");
			println("delete cb;");
			println("cb = NULL;");
			println("}");
		}
		println("}");
		println("");
	}


	/**
	 * HELPER: writeCplusMethodCallbackPermission() writes permission checks in stub for callbacks
	 */
	private void writeCplusMethodCallbackPermission(String intface) {

		println("int methodId = IoTRMIObject::getMethodId(method);");
		// Get all the different stubs
		Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
		for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {
			String newIntface = intMeth.getKey();
			int newObjectId = mapNewIntfaceObjId.get(newIntface);
			println("if (set" + newObjectId + "Allowed.find(methodId) == set" + newObjectId + "Allowed.end()) {");
			println("cerr << \"Callback object for " + intface + " is not allowed to access method: \" << methodId;");
			println("exit(-1);");
			println("}");
		}
	}


	/**
	 * HELPER: writeInitCallbackCplusStub() writes the initialization of callback
	 */
	private void writeInitCallbackCplusStub(String intface, InterfaceDecl intDecl) {

		println("void ___initCallBack() {");
		println("bool bResult = false;");
		println("rmiObj = new IoTRMIObject(ports[0], &bResult);");
		println("while (true) {");
		println("char* method = rmiObj->getMethodBytes();");
		writeCplusMethodCallbackPermission(intface);
		println("int objId = IoTRMIObject::getObjectId(method);");
		println("if (objId < vecCallbackObj.size()) {	// Check if still within range");
		println(intface + "_CallbackSkeleton* skel = dynamic_cast<" + intface + 
			"_CallbackSkeleton*> (vecCallbackObj.at(objId));");
		println("skel->invokeMethod(rmiObj);");
		println("} else {");
		println("cerr << \"Illegal object Id: \" << to_string(objId);");
		// TODO: perhaps need to change this into "throw" to make it cleaner (allow stack unfolding)
		println("exit(-1);");
		println("}");
		println("}");
		println("}\n");
	}


	/**
	 * HELPER: writeInitCallbackSendInfoCplusStub() writes the initialization of callback
	 */
	private void writeInitCallbackSendInfoCplusStub(InterfaceDecl intDecl) {

		// Generate info sending part
		println("void ___regCB() {");
		println("int numParam = 3;");
		String method = "___initCallBack()";
		println("int methodId = " + intDecl.getHelperMethodNumId(method) + ";");
		println("string retType = \"void\";");
		println("string paramCls[] = { \"int\", \"string\", \"int\" };");
		println("int rev = 0;");
		println("void* paramObj[] = { &ports[0], &address, &rev };");
		println("void* retObj = NULL;");
		println("rmiCall->remoteCall(objectId, methodId, retType, paramCls, paramObj, numParam, retObj);");
		println("}\n");
	}


	/**
	 * generateCPlusStubClasses() generate stubs based on the methods list in C++
	 */
	public void generateCPlusStubClasses() throws IOException {

		// Create a new directory
		String path = createDirectories(dir, subdir);
		for (String intface : mapIntfacePTH.keySet()) {

			Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
			for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {
				// Open a new file to write into
				String newIntface = intMeth.getKey();
				String newStubClass = newIntface + "_Stub";
				FileWriter fw = new FileWriter(path + "/" + newStubClass + ".hpp");
				pw = new PrintWriter(new BufferedWriter(fw));
				// Write file headers
				println("#ifndef _" + newStubClass.toUpperCase() + "_HPP__");
				println("#define _" + newStubClass.toUpperCase() + "_HPP__");
				println("#include <iostream>");
				// Find out if there are callback objects
				Set<String> methods = intMeth.getValue();
				DeclarationHandler decHandler = mapIntDeclHand.get(intface);
				InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
				Set<String> callbackClasses = getCallbackClasses(methods, intDecl);
				boolean callbackExist = !callbackClasses.isEmpty();
				if (callbackExist)	// Need thread library if this has callback
					println("#include <thread>");
				println("#include \"" + newIntface + ".hpp\""); println("");		
				println("using namespace std;"); println("");
				println("class " + newStubClass + " : public " + newIntface); println("{");
				println("private:\n");
				writePropertiesCplusStub(intface, newIntface, callbackExist, callbackClasses);
				println("public:\n");
				// Add default constructor and destructor
				println(newStubClass + "() { }"); println("");
				writeConstructorCplusStub(newStubClass, callbackExist, callbackClasses);
				writeDeconstructorCplusStub(newStubClass, callbackExist, callbackClasses);
				// Write methods
				writeMethodCplusStub(methods, intDecl, callbackClasses);
				print("}"); println(";");
				if (callbackExist)
					writePermissionInitializationCplus(intface, newStubClass, intDecl);
				println("#endif");
				pw.close();
				System.out.println("IoTCompiler: Generated stub class " + newStubClass + ".hpp...");
			}
		}
	}


	/**
	 * HELPER: writePropertiesCplusCallbackStub() writes the properties of the stub class
	 */
	private void writePropertiesCplusCallbackStub(String intface, String newIntface, boolean callbackExist, Set<String> callbackClasses) {

		println("IoTRMICall *rmiCall;");
		// Get the object Id
		println("static int objectId;");
		if (callbackExist) {
		// We assume that each class only has one callback interface for now
			Iterator it = callbackClasses.iterator();
			String callbackType = (String) it.next();
			println("// Callback properties");
			println("IoTRMIObject *rmiObj;");
			println("vector<" + callbackType + "*> vecCallbackObj;");
			println("static int objIdCnt;");
			// TODO: Need to initialize address and ports if we want to have callback-in-callback
			println("string address;");
			println("vector<int> ports;\n");
			writePropertiesCplusPermission(callbackType);
		}
		println("\n");
	}


	/**
	 * HELPER: writeConstructorCplusCallbackStub() writes the constructor of the stub class
	 */
	private void writeConstructorCplusCallbackStub(String newStubClass, boolean callbackExist, Set<String> callbackClasses) {

		println(newStubClass + "(IoTRMICall* _rmiCall, int _objectId) {");
		println("objectId = _objectId;");
		println("rmiCall = _rmiCall;");
		if (callbackExist) {
			Iterator it = callbackClasses.iterator();
			String callbackType = (String) it.next();
			println("thread th1 (&" + newStubClass + "::___initCallBack, this);");
			println("th1.detach();");
			println("___regCB();");
		}
		println("}\n");
	}


	/**
	 * generateCPlusCallbackStubClasses() generate callback stubs based on the methods list in C++
	 */
	public void generateCPlusCallbackStubClasses() throws IOException {

		// Create a new directory
		String path = createDirectories(dir, subdir);
		for (String intface : mapIntfacePTH.keySet()) {

			Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
			for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {
				// Open a new file to write into
				String newIntface = intMeth.getKey();
				String newStubClass = newIntface + "_CallbackStub";
				FileWriter fw = new FileWriter(path + "/" + newStubClass + ".hpp");
				pw = new PrintWriter(new BufferedWriter(fw));
				// Find out if there are callback objects
				Set<String> methods = intMeth.getValue();
				DeclarationHandler decHandler = mapIntDeclHand.get(intface);
				InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
				Set<String> callbackClasses = getCallbackClasses(methods, intDecl);
				boolean callbackExist = !callbackClasses.isEmpty();
				// Write file headers
				println("#ifndef _" + newStubClass.toUpperCase() + "_HPP__");
				println("#define _" + newStubClass.toUpperCase() + "_HPP__");
				println("#include <iostream>");
				if (callbackExist)
					println("#include <thread>");
				println("#include \"" + newIntface + ".hpp\""); println("");		
				println("using namespace std;"); println("");
				println("class " + newStubClass + " : public " + newIntface); println("{");
				println("private:\n");
				writePropertiesCplusCallbackStub(intface, newIntface, callbackExist, callbackClasses);
				println("public:\n");
				// Add default constructor and destructor
				println(newStubClass + "() { }"); println("");
				writeConstructorCplusCallbackStub(newStubClass, callbackExist, callbackClasses);
				writeDeconstructorCplusStub(newStubClass, callbackExist, callbackClasses);
				// Write methods
				writeMethodCplusStub(methods, intDecl, callbackClasses);
				println("};");
				if (callbackExist)
					writePermissionInitializationCplus(intface, newStubClass, intDecl);
				println("#endif");
				pw.close();
				System.out.println("IoTCompiler: Generated callback stub class " + newIntface + ".hpp...");
			}
		}
	}


	/**
	 * HELPER: writePropertiesCplusSkeleton() writes the properties of the skeleton class
	 */
	private void writePropertiesCplusSkeleton(String intface, boolean callbackExist, Set<String> callbackClasses) {

		println(intface + " *mainObj;");
		// Callback
		if (callbackExist) {
			Iterator it = callbackClasses.iterator();
			String callbackType = (String) it.next();
			String exchangeType = checkAndGetParamClass(callbackType);
			println("// Callback properties");
			println("static int objIdCnt;");
			println("vector<" + exchangeType + "*> vecCallbackObj;");
			println("IoTRMICall *rmiCall;");
		}
		println("IoTRMIObject *rmiObj;\n");
		// Keep track of object Ids of all stubs registered to this interface
		writePropertiesCplusPermission(intface);
		println("\n");
	}


	/**
	 * HELPER: writePermissionInitializationCplus() writes the initialization of permission set
	 */
	private void writePermissionInitializationCplus(String intface, String newSkelClass, InterfaceDecl intDecl) {

		// Keep track of object Ids of all stubs registered to this interface
		Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
		for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {
			String newIntface = intMeth.getKey();
			int newObjectId = mapNewIntfaceObjId.get(newIntface);
			print("const set<int> " + newSkelClass + "::set" + newObjectId + "Allowed {");
			Set<String> methodIds = intMeth.getValue();
			int i = 0;
			for (String methodId : methodIds) {
				int methodNumId = intDecl.getMethodNumId(methodId);
				print(Integer.toString(methodNumId));
				// Check if this is the last element (don't print a comma)
				if (i != methodIds.size() - 1) {
					print(", ");
				}
				i++;
			}
			println(" };");
		}	
	}


	/**
	 * HELPER: writeConstructorCplusSkeleton() writes the constructor of the skeleton class
	 */
	private void writeConstructorCplusSkeleton(String newSkelClass, String intface, boolean callbackExist) {

		println(newSkelClass + "(" + intface + " *_mainObj, int _port) {");
		println("bool _bResult = false;");
		println("mainObj = _mainObj;");
		println("rmiObj = new IoTRMIObject(_port, &_bResult);");
		// Callback
		if (callbackExist) {
			println("objIdCnt = 0;");
		}
		//println("set0Allowed = Arrays.asList(object0Permission);");
		println("___waitRequestInvokeMethod();");
		println("}\n");
	}


	/**
	 * HELPER: writeDeconstructorCplusSkeleton() writes the deconstructor of the skeleton class
	 */
	private void writeDeconstructorCplusSkeleton(String newSkelClass, boolean callbackExist, Set<String> callbackClasses) {

		println("~" + newSkelClass + "() {");
		println("if (rmiObj != NULL) {");
		println("delete rmiObj;");
		println("rmiObj = NULL;");
		println("}");
		if (callbackExist) {
		// We assume that each class only has one callback interface for now
			println("if (rmiCall != NULL) {");
			println("delete rmiCall;");
			println("rmiCall = NULL;");
			println("}");
			Iterator it = callbackClasses.iterator();
			String callbackType = (String) it.next();
			String exchangeType = checkAndGetParamClass(callbackType);
			println("for(" + exchangeType + "* cb : vecCallbackObj) {");
			println("delete cb;");
			println("cb = NULL;");
			println("}");
		}
		println("}");
		println("");
	}


	/**
	 * HELPER: writeStdMethodBodyCplusSkeleton() writes the standard method body in the skeleton class
	 */
	private void writeStdMethodBodyCplusSkeleton(List<String> methParams, String methodId, String methodType) {

		if (methodType.equals("void"))
			print("mainObj->" + methodId + "(");
		else
			print("return mainObj->" + methodId + "(");
		for (int i = 0; i < methParams.size(); i++) {

			print(getSimpleIdentifier(methParams.get(i)));
			// Check if this is the last element (don't print a comma)
			if (i != methParams.size() - 1) {
				print(", ");
			}
		}
		println(");");
	}


	/**
	 * HELPER: writeInitCallbackCplusSkeleton() writes the init callback method for skeleton class
	 */
	private void writeInitCallbackCplusSkeleton(boolean callbackSkeleton) {

		// This is a callback skeleton generation
		if (callbackSkeleton)
			println("void ___regCB(IoTRMIObject* rmiObj) {");
		else
			println("void ___regCB() {");
		println("int numParam = 3;");
		println("int param1 = 0;");
		println("string param2 = \"\";");
		println("int param3 = 0;");
		println("void* paramObj[] = { &param1, &param2, &param3 };");
		println("bool bResult = false;");
		println("rmiCall = new IoTRMICall(param1, param2.c_str(), param3, &bResult);");
		println("}\n");
	}


	/**
	 * HELPER: writeMethodCplusSkeleton() writes the method of the skeleton class
	 */
	private void writeMethodCplusSkeleton(Collection<String> methods, InterfaceDecl intDecl, 
			Set<String> callbackClasses, boolean callbackSkeleton) {

		for (String method : methods) {

			List<String> methParams = intDecl.getMethodParams(method);
			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			String methodId = intDecl.getMethodId(method);
			String methodType = checkAndGetCplusType(intDecl.getMethodType(method));
			print(methodType + " " + methodId + "(");
			boolean isCallbackMethod = false;
			String callbackType = null;
			for (int i = 0; i < methParams.size(); i++) {

				String origParamType = methPrmTypes.get(i);
				if (callbackClasses.contains(origParamType)) { // Check if this has callback object
					isCallbackMethod = true;
					callbackType = origParamType;	
				}
				String paramType = checkAndGetParamClass(methPrmTypes.get(i));
				String methPrmType = checkAndGetCplusType(paramType);
				String methParamComplete = checkAndGetCplusArray(methPrmType, methParams.get(i));
				print(methParamComplete);
				// Check if this is the last element (don't print a comma)
				if (i != methParams.size() - 1) {
					print(", ");
				}
			}
			println(") {");
			// Now, write the body of skeleton!
			writeStdMethodBodyCplusSkeleton(methParams, methodId, intDecl.getMethodType(method));
			println("}\n");
			if (isCallbackMethod)
				writeInitCallbackCplusSkeleton(callbackSkeleton);
		}
	}


	/**
	 * HELPER: writeCallbackCplusNumStubs() writes the numStubs variable
	 */
	private void writeCallbackCplusNumStubs(List<String> methParams, List<String> methPrmTypes, String callbackType) {

		for (int i = 0; i < methParams.size(); i++) {
			String paramType = methPrmTypes.get(i);
			String param = methParams.get(i);
			//if (callbackType.equals(paramType)) {
			if (checkCallbackType(paramType, callbackType)) { // Check if this has callback object
				String exchParamType = checkAndGetParamClass(paramType);
				// Print array if this is array or list if this is a list of callback objects
				println("int numStubs" + i + " = 0;");
			}
		}
	}


	/**
	 * HELPER: writeCallbackCplusStubGeneration() writes the callback stub generation part
	 */
	private void writeCallbackCplusStubGeneration(List<String> methParams, List<String> methPrmTypes, String callbackType) {

		// Iterate over callback objects
		for (int i = 0; i < methParams.size(); i++) {
			String paramType = methPrmTypes.get(i);
			String param = methParams.get(i);
			// Generate a loop if needed
			if (checkCallbackType(paramType, callbackType)) { // Check if this has callback object
				String exchParamType = checkAndGetParamClass(paramType);
				if (isArrayOrList(paramType, param)) {
					println("vector<" + exchParamType + "> stub;");
					println("for (int objId = 0; objId < numStubs" + i + "; objId++) {");
					println(exchParamType + "* cb" + i + " = new " + exchParamType + "_CallbackStub(rmiCall, objIdCnt);");
					println("stub" + i + ".push_back(cb);");
					println("vecCallbackObj.push_back(cb);");
					println("objIdCnt++;");
					println("}");
				} else {
					println(exchParamType + "* stub" + i + " = new " + exchParamType + "_CallbackStub(rmiCall, objIdCnt);");
					println("vecCallbackObj.push_back(stub" + i + ");");
					println("objIdCnt++;");
				}
			}
		}
	}


	/**
	 * HELPER: writeStdMethodHelperBodyCplusSkeleton() writes the standard method body helper in the skeleton class
	 */
	private void writeStdMethodHelperBodyCplusSkeleton(InterfaceDecl intDecl, List<String> methParams,
			List<String> methPrmTypes, String method, String methodId, Set<String> callbackClasses) {

		// Generate array of parameter types
		boolean isCallbackMethod = false;
		String callbackType = null;
		print("string paramCls[] = { ");
		for (int i = 0; i < methParams.size(); i++) {
			String paramType = returnGenericCallbackType(methPrmTypes.get(i));
			if (callbackClasses.contains(paramType)) {
				isCallbackMethod = true;
				callbackType = paramType;
				print("\"int\"");
			} else {	// Generate normal classes if it's not a callback object
				String paramTypeC = checkAndGetCplusType(methPrmTypes.get(i));
				String prmType = checkAndGetCplusArrayType(paramTypeC, methParams.get(i));
				print("\"" + prmType + "\"");
			}
			if (i != methParams.size() - 1) {
				print(", ");
			}
		}
		println(" };");
		println("int numParam = " + methParams.size() + ";");
		if (isCallbackMethod)
			writeCallbackCplusNumStubs(methParams, methPrmTypes, callbackType);
		// Generate parameters
		for (int i = 0; i < methParams.size(); i++) {
			String paramType = returnGenericCallbackType(methPrmTypes.get(i));
			if (!callbackClasses.contains(paramType)) {			
				String methPrmType = checkAndGetCplusType(methPrmTypes.get(i));
				String methParamComplete = checkAndGetCplusArray(methPrmType, methParams.get(i));
				println(methParamComplete + ";");
			}
		}
		// Generate array of parameter objects
		print("void* paramObj[] = { ");
		for (int i = 0; i < methParams.size(); i++) {
			String paramType = returnGenericCallbackType(methPrmTypes.get(i));
			if (callbackClasses.contains(paramType))
				print("&numStubs" + i);
			else
				print("&" + getSimpleIdentifier(methParams.get(i)));
			if (i != methParams.size() - 1) {
				print(", ");
			}
		}
		println(" };");
		println("rmiObj->getMethodParams(paramCls, numParam, paramObj);");
		if (isCallbackMethod)
			writeCallbackCplusStubGeneration(methParams, methPrmTypes, callbackType);
		String retType = intDecl.getMethodType(method);
		// Check if this is "void"
		if (retType.equals("void")) {
			print(methodId + "(");
			for (int i = 0; i < methParams.size(); i++) {
				String paramType = returnGenericCallbackType(methPrmTypes.get(i));
				if (callbackClasses.contains(paramType))
					print("stub" + i);
				else
					print(getSimpleIdentifier(methParams.get(i)));
				if (i != methParams.size() - 1) {
					print(", ");
				}
			}
			println(");");
		} else { // We do have a return value
			print(checkAndGetCplusType(retType) + " retVal = " + methodId + "(");
			for (int i = 0; i < methParams.size(); i++) {
				String paramType = returnGenericCallbackType(methPrmTypes.get(i));
				if (callbackClasses.contains(paramType))
					print("stub" + i);
				else
					print(getSimpleIdentifier(methParams.get(i)));
				if (i != methParams.size() - 1) {
					print(", ");
				}
			}
			println(");");
			println("void* retObj = &retVal;");
			String retTypeC = checkAndGetCplusType(retType);
			println("rmiObj->sendReturnObj(retObj, \"" + checkAndGetCplusArrayType(retTypeC) + "\");");
		}
	}


	/**
	 * HELPER: writeMethodHelperCplusSkeleton() writes the method helper of the skeleton class
	 */
	private void writeMethodHelperCplusSkeleton(Collection<String> methods, InterfaceDecl intDecl, Set<String> callbackClasses) {

		// Use this set to handle two same methodIds
		Set<String> uniqueMethodIds = new HashSet<String>();
		for (String method : methods) {

			List<String> methParams = intDecl.getMethodParams(method);
			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			String methodId = intDecl.getMethodId(method);
			print("void ___");
			String helperMethod = methodId;
			if (uniqueMethodIds.contains(methodId))
				helperMethod = helperMethod + intDecl.getMethodNumId(method);
			else
				uniqueMethodIds.add(methodId);
			// Check if this is "void"
			String retType = intDecl.getMethodType(method);
			println(helperMethod + "() {");
			// Now, write the helper body of skeleton!
			writeStdMethodHelperBodyCplusSkeleton(intDecl, methParams, methPrmTypes, method, methodId, callbackClasses);
			println("}\n");
		}
	}


	/**
	 * HELPER: writeCplusMethodPermission() writes permission checks in skeleton
	 */
	private void writeCplusMethodPermission(String intface) {

		// Get all the different stubs
		Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
		for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {
			String newIntface = intMeth.getKey();
			int newObjectId = mapNewIntfaceObjId.get(newIntface);
			println("if (_objectId == object" + newObjectId + "Id) {");
			println("if (set" + newObjectId + "Allowed.find(methodId) == set" + newObjectId + "Allowed.end()) {");
			println("cerr << \"Object with object Id: \" << _objectId << \"  is not allowed to access method: \" << methodId << endl;");
			println("exit(-1);");
			println("}");
			println("else {");
			println("cerr << \"Object Id: \" << _objectId << \" not recognized!\" << endl;");
			println("exit(-1);");
			println("}");
			println("}");
		}
	}


	/**
	 * HELPER: writeCplusWaitRequestInvokeMethod() writes the main loop of the skeleton class
	 */
	private void writeCplusWaitRequestInvokeMethod(Collection<String> methods, InterfaceDecl intDecl, boolean callbackExist, String intface) {

		// Use this set to handle two same methodIds
		Set<String> uniqueMethodIds = new HashSet<String>();
		println("void ___waitRequestInvokeMethod() {");
		// Write variables here if we have callbacks or enums or structs
		println("while (true) {");
		println("rmiObj->getMethodBytes();");
		println("int _objectId = rmiObj->getObjectId();");
		println("int methodId = rmiObj->getMethodId();");
		// Generate permission check
		writeCplusMethodPermission(intface);
		println("switch (methodId) {");
		// Print methods and method Ids
		for (String method : methods) {
			String methodId = intDecl.getMethodId(method);
			int methodNumId = intDecl.getMethodNumId(method);
			print("case " + methodNumId + ": ___");
			String helperMethod = methodId;
			if (uniqueMethodIds.contains(methodId))
				helperMethod = helperMethod + methodNumId;
			else
				uniqueMethodIds.add(methodId);
			println(helperMethod + "(); break;");
		}
		String method = "___initCallBack()";
		// Print case -9999 (callback handler) if callback exists
		if (callbackExist) {
			int methodId = intDecl.getHelperMethodNumId(method);
			println("case " + methodId + ": ___regCB(); break;");
		}
		println("default: ");
		println("cerr << \"Method Id \" << methodId << \" not recognized!\" << endl;");
		println("throw exception();");
		println("}");
		println("}");
		println("}\n");
	}


	/**
	 * generateCplusSkeletonClass() generate skeletons based on the methods list in C++
	 */
	public void generateCplusSkeletonClass() throws IOException {

		// Create a new directory
		String path = createDirectories(dir, subdir);
		for (String intface : mapIntfacePTH.keySet()) {
			// Open a new file to write into
			String newSkelClass = intface + "_Skeleton";
			FileWriter fw = new FileWriter(path + "/" + newSkelClass + ".hpp");
			pw = new PrintWriter(new BufferedWriter(fw));
			// Write file headers
			println("#ifndef _" + newSkelClass.toUpperCase() + "_HPP__");
			println("#define _" + newSkelClass.toUpperCase() + "_HPP__");
			println("#include <iostream>");
			println("#include \"" + intface + ".hpp\"\n");
			// Pass in set of methods and get import classes
			DeclarationHandler decHandler = mapIntDeclHand.get(intface);
			InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
			List<String> methods = intDecl.getMethods();
			Set<String> includeClasses = getIncludeClasses(methods, intDecl, true);
			List<String> stdIncludeClasses = getStandardCplusIncludeClasses();
			List<String> allIncludeClasses = getAllLibClasses(stdIncludeClasses, includeClasses);
			printIncludeStatements(allIncludeClasses); println("");
			println("using namespace std;\n");
			// Find out if there are callback objects
			Set<String> callbackClasses = getCallbackClasses(methods, intDecl);
			boolean callbackExist = !callbackClasses.isEmpty();
			// Write class header
			println("class " + newSkelClass + " : public " + intface); println("{");
			println("private:\n");
			// Write properties
			writePropertiesCplusSkeleton(intface, callbackExist, callbackClasses);
			println("public:\n");
			// Write constructor
			writeConstructorCplusSkeleton(newSkelClass, intface, callbackExist);
			// Write deconstructor
			writeDeconstructorCplusSkeleton(newSkelClass, callbackExist, callbackClasses);
			// Write methods
			writeMethodCplusSkeleton(methods, intDecl, callbackClasses, false);
			// Write method helper
			writeMethodHelperCplusSkeleton(methods, intDecl, callbackClasses);
			// Write waitRequestInvokeMethod() - main loop
			writeCplusWaitRequestInvokeMethod(methods, intDecl, callbackExist, intface);
			println("};");
			writePermissionInitializationCplus(intface, newSkelClass, intDecl);
			println("#endif");
			pw.close();
			System.out.println("IoTCompiler: Generated skeleton class " + newSkelClass + ".hpp...");
		}
	}


	/**
	 * HELPER: writePropertiesCplusCallbackSkeleton() writes the properties of the callback skeleton class
	 */
	private void writePropertiesCplusCallbackSkeleton(String intface, boolean callbackExist, Set<String> callbackClasses) {

		println(intface + " *mainObj;");
		// Keep track of object Ids of all stubs registered to this interface
		println("static int objectId;");
		// Callback
		if (callbackExist) {
			Iterator it = callbackClasses.iterator();
			String callbackType = (String) it.next();
			String exchangeType = checkAndGetParamClass(callbackType);
			println("// Callback properties");
			println("IoTRMICall* rmiCall;");
			println("vector<" + exchangeType + "*> vecCallbackObj;");
			println("static int objIdCnt;");
		}
		println("\n");
	}


	/**
	 * HELPER: writeConstructorCplusCallbackSkeleton() writes the constructor of the skeleton class
	 */
	private void writeConstructorCplusCallbackSkeleton(String newSkelClass, String intface, boolean callbackExist) {

		println(newSkelClass + "(" + intface + " *_mainObj, int _objectId) {");
		println("mainObj = _mainObj;");
		println("objectId = _objectId;");
		// Callback
		if (callbackExist) {
			println("objIdCnt = 0;");
		}
		println("}\n");
	}


	/**
	 * HELPER: writeDeconstructorCplusStub() writes the deconstructor of the stub class
	 */
	private void writeDeconstructorCplusCallbackSkeleton(String newStubClass, boolean callbackExist, 
			Set<String> callbackClasses) {

		println("~" + newStubClass + "() {");
		if (callbackExist) {
		// We assume that each class only has one callback interface for now
			println("if (rmiCall != NULL) {");
			println("delete rmiCall;");
			println("rmiCall = NULL;");
			println("}");
			Iterator it = callbackClasses.iterator();
			String callbackType = (String) it.next();
			String exchangeType = checkAndGetParamClass(callbackType);
			println("for(" + exchangeType + "* cb : vecCallbackObj) {");
			println("delete cb;");
			println("cb = NULL;");
			println("}");
		}
		println("}");
		println("");
	}


	/**
	 * HELPER: writeMethodHelperCplusCallbackSkeleton() writes the method helper of the callback skeleton class
	 */
	private void writeMethodHelperCplusCallbackSkeleton(Collection<String> methods, InterfaceDecl intDecl, 
			Set<String> callbackClasses) {

		// Use this set to handle two same methodIds
		Set<String> uniqueMethodIds = new HashSet<String>();
		for (String method : methods) {

			List<String> methParams = intDecl.getMethodParams(method);
			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			String methodId = intDecl.getMethodId(method);
			print("void ___");
			String helperMethod = methodId;
			if (uniqueMethodIds.contains(methodId))
				helperMethod = helperMethod + intDecl.getMethodNumId(method);
			else
				uniqueMethodIds.add(methodId);
			// Check if this is "void"
			String retType = intDecl.getMethodType(method);
			println(helperMethod + "(IoTRMIObject* rmiObj) {");
			// Now, write the helper body of skeleton!
			writeStdMethodHelperBodyCplusSkeleton(intDecl, methParams, methPrmTypes, method, methodId, callbackClasses);
			println("}\n");
		}
	}


	/**
	 * HELPER: writeCplusCallbackWaitRequestInvokeMethod() writes the main loop of the skeleton class
	 */
	private void writeCplusCallbackWaitRequestInvokeMethod(Collection<String> methods, InterfaceDecl intDecl, 
			boolean callbackExist) {

		// Use this set to handle two same methodIds
		Set<String> uniqueMethodIds = new HashSet<String>();
		println("void invokeMethod(IoTRMIObject* rmiObj) {");
		// Write variables here if we have callbacks or enums or structs
		println("int methodId = rmiObj->getMethodId();");
		// TODO: code the permission check here!
		println("switch (methodId) {");
		// Print methods and method Ids
		for (String method : methods) {
			String methodId = intDecl.getMethodId(method);
			int methodNumId = intDecl.getMethodNumId(method);
			print("case " + methodNumId + ": ___");
			String helperMethod = methodId;
			if (uniqueMethodIds.contains(methodId))
				helperMethod = helperMethod + methodNumId;
			else
				uniqueMethodIds.add(methodId);
			println(helperMethod + "(rmiObj); break;");
		}
		String method = "___initCallBack()";
		// Print case -9999 (callback handler) if callback exists
		if (callbackExist) {
			int methodId = intDecl.getHelperMethodNumId(method);
			println("case " + methodId + ": ___regCB(rmiObj); break;");
		}
		println("default: ");
		println("cerr << \"Method Id \" << methodId << \" not recognized!\" << endl;");
		println("throw exception();");
		println("}");
		println("}\n");
	}



	/**
	 * generateCplusCallbackSkeletonClass() generate callback skeletons based on the methods list in C++
	 */
	public void generateCplusCallbackSkeletonClass() throws IOException {

		// Create a new directory
		String path = createDirectories(dir, subdir);
		for (String intface : mapIntfacePTH.keySet()) {
			// Open a new file to write into
			String newSkelClass = intface + "_CallbackSkeleton";
			FileWriter fw = new FileWriter(path + "/" + newSkelClass + ".hpp");
			pw = new PrintWriter(new BufferedWriter(fw));
			// Write file headers
			println("#ifndef _" + newSkelClass.toUpperCase() + "_HPP__");
			println("#define _" + newSkelClass.toUpperCase() + "_HPP__");
			println("#include <iostream>");
			println("#include \"" + intface + ".hpp\"\n");
			// Pass in set of methods and get import classes
			DeclarationHandler decHandler = mapIntDeclHand.get(intface);
			InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
			List<String> methods = intDecl.getMethods();
			Set<String> includeClasses = getIncludeClasses(methods, intDecl, true);
			List<String> stdIncludeClasses = getStandardCplusIncludeClasses();
			List<String> allIncludeClasses = getAllLibClasses(stdIncludeClasses, includeClasses);
			printIncludeStatements(allIncludeClasses); println("");			
			// Find out if there are callback objects
			Set<String> callbackClasses = getCallbackClasses(methods, intDecl);
			boolean callbackExist = !callbackClasses.isEmpty();
			println("using namespace std;\n");
			// Write class header
			println("class " + newSkelClass + " : public " + intface); println("{");
			println("private:\n");
			// Write properties
			writePropertiesCplusCallbackSkeleton(intface, callbackExist, callbackClasses);
			println("public:\n");
			// Write constructor
			writeConstructorCplusCallbackSkeleton(newSkelClass, intface, callbackExist);
			// Write deconstructor
			writeDeconstructorCplusCallbackSkeleton(newSkelClass, callbackExist, callbackClasses);
			// Write methods
			writeMethodCplusSkeleton(methods, intDecl, callbackClasses, true);
			// Write method helper
			writeMethodHelperCplusCallbackSkeleton(methods, intDecl, callbackClasses);
			// Write waitRequestInvokeMethod() - main loop
			writeCplusCallbackWaitRequestInvokeMethod(methods, intDecl, callbackExist);
			println("};");
			println("#endif");
			pw.close();
			System.out.println("IoTCompiler: Generated callback skeleton class " + newSkelClass + ".hpp...");
		}
	}


	/**
	 * generateInitializer() generate initializer based on type
	 */
	public String generateCplusInitializer(String type) {

		// Generate dummy returns for now
		if (type.equals("short")||
			type.equals("int") 	||
			type.equals("long") ||
			type.equals("float")||
			type.equals("double")) {

			return "0";
		} else if ( type.equals("String") ||
					type.equals("string")) {
  
			return "\"\"";
		} else if ( type.equals("char") ||
					type.equals("byte")) {

			return "\' \'";
		} else if ( type.equals("boolean")) {

			return "false";
		} else {
			return "NULL";
		}
	}


	/**
	 * generateReturnStmt() generate return statement based on methType
	 */
	public String generateReturnStmt(String methType) {

		// Generate dummy returns for now
		if (methType.equals("short")||
			methType.equals("int") 	||
			methType.equals("long") ||
			methType.equals("float")||
			methType.equals("double")) {

			return "1";
		} else if ( methType.equals("String")) {
  
			return "\"a\"";
		} else if ( methType.equals("char") ||
					methType.equals("byte")) {

			return "\'a\'";
		} else if ( methType.equals("boolean")) {

			return "true";
		} else {
			return "null";
		}
	}


	/**
	 * setDirectory() sets a new directory for stub files
	 */
	public void setDirectory(String _subdir) {

		subdir = _subdir;
	}


	/**
	 * printUsage() prints the usage of this compiler
	 */
	public static void printUsage() {

		System.out.println();
		System.out.println("Sentinel interface and stub compiler version 1.0");
		System.out.println("Copyright (c) 2015-2016 University of California, Irvine - Programming Language Group.");
		System.out.println("All rights reserved.");
		System.out.println("Usage:");
		System.out.println("\tjava IoTCompiler -help / --help / -h\n");
		System.out.println("\t\tDisplay this help texts\n\n");
		System.out.println("\tjava IoTCompiler [<main-policy-file> <req-policy-file>]");
		System.out.println("\tjava IoTCompiler [<main-policy-file> <req-policy-file>] [options]\n");
		System.out.println("\t\tTake one or more pairs of main-req policy files, and generate Java and/or C++ files\n");
		System.out.println("Options:");
		System.out.println("\t-java\t<directory>\tGenerate Java stub files");
		System.out.println("\t-cplus\t<directory>\tGenerate C++ stub files");
		System.out.println();
	}


	/**
	 * parseFile() prepares Lexer and Parser objects, then parses the file
	 */
	public static ParseNode parseFile(String file) {

		ParseNode pn = null;
		try {
			ComplexSymbolFactory csf = new ComplexSymbolFactory();
			ScannerBuffer lexer = 
				new ScannerBuffer(new Lexer(new BufferedReader(new FileReader(file)),csf));
			Parser parse = new Parser(lexer,csf);
			pn = (ParseNode) parse.parse().value;
		} catch (Exception e) {
			e.printStackTrace();
			throw new Error("IoTCompiler: ERROR parsing policy file or wrong command line option: " + file);
		}

		return pn;
	}


	/**================
	 * Helper functions
	 **================
	 */
	boolean newline=true;
	int tablevel=0;

	private void print(String str) {
		if (newline) {
			int tab=tablevel;
			if (str.equals("}"))
				tab--;
			for(int i=0; i<tab; i++)
				pw.print("\t");
		}
		pw.print(str);
		updatetabbing(str);
		newline=false;
	}


	/**
	 * This function converts Java to C++ type for compilation
	 */
	private String convertType(String jType) {

		return mapPrimitives.get(jType);
	}


	private void println(String str) {
		if (newline) {
			int tab = tablevel;
			if (str.contains("}") && !str.contains("{"))
				tab--;
			for(int i=0; i<tab; i++)
				pw.print("\t");
		}
		pw.println(str);
		updatetabbing(str);
		newline = true;
	}


	private void updatetabbing(String str) {

		tablevel+=count(str,'{')-count(str,'}');
	}


	private int count(String str, char key) {
		char[] array = str.toCharArray();
		int count = 0;
		for(int i=0; i<array.length; i++) {
			if (array[i] == key)
				count++;
		}
		return count;
	}


	private void createDirectory(String dirName) {

		File file = new File(dirName);
		if (!file.exists()) {
			if (file.mkdir()) {
				System.out.println("IoTCompiler: Directory " + dirName + " has been created!");
			} else {
				System.out.println("IoTCompiler: Failed to create directory " + dirName + "!");
			}
		} else {
			System.out.println("IoTCompiler: Directory " + dirName + " exists...");
		}
	}


	// Create a directory and possibly a sub directory
	private String createDirectories(String dir, String subdir) {

		String path = dir;
		createDirectory(path);
		if (subdir != null) {
			path = path + "/" + subdir;
			createDirectory(path);
		}
		return path;
	}


	// Inserting array members into a Map object
	// that maps arrKey to arrVal objects
	private void arraysToMap(Map map, Object[] arrKey, Object[] arrVal) {

		for(int i = 0; i < arrKey.length; i++) {

			map.put(arrKey[i], arrVal[i]);
		}
	}


	// Return parameter category, i.e. PRIMITIVES, NONPRIMITIVES, or USERDEFINED
	private ParamCategory getParamCategory(String paramType) {

		if (mapPrimitives.containsKey(paramType)) {
			return ParamCategory.PRIMITIVES;
		// We can either use mapNonPrimitivesJava or mapNonPrimitivesCplus here
		} else if (mapNonPrimitivesJava.containsKey(getSimpleType(paramType))) {
			return ParamCategory.NONPRIMITIVES;
		} else
			return ParamCategory.USERDEFINED;
	}


	// Return full class name for non-primitives to generate Java import statements
	// e.g. java.util.Set for Set, java.util.Map for Map
	private String getNonPrimitiveJavaClass(String paramNonPrimitives) {

		return mapNonPrimitivesJava.get(paramNonPrimitives);
	}


	// Return full class name for non-primitives to generate Cplus include statements
	// e.g. #include <set> for Set, #include <map> for Map
	private String getNonPrimitiveCplusClass(String paramNonPrimitives) {

		return mapNonPrimitivesCplus.get(paramNonPrimitives);
	}


	// Get simple types, e.g. HashSet for HashSet<...>
	// Basically strip off the "<...>"
	private String getSimpleType(String paramType) {

		// Check if this is generics
		if(paramType.contains("<")) {
			String[] type = paramType.split("<");
			return type[0];
		} else
			return paramType;
	}


	// Generate a set of standard classes for import statements
	private List<String> getStandardJavaImportClasses() {

		List<String> importClasses = new ArrayList<String>();
		// Add the standard list first
		importClasses.add("java.io.IOException");
		importClasses.add("java.util.List");
		importClasses.add("java.util.ArrayList");
		importClasses.add("java.util.Arrays");
		importClasses.add("iotrmi.Java.IoTRMICall");
		importClasses.add("iotrmi.Java.IoTRMIObject");

		return importClasses;
	}


	// Generate a set of standard classes for import statements
	private List<String> getStandardCplusIncludeClasses() {

		List<String> importClasses = new ArrayList<String>();
		// Add the standard list first
		importClasses.add("<vector>");
		importClasses.add("<set>");
		importClasses.add("\"IoTRMICall.hpp\"");
		importClasses.add("\"IoTRMIObject.hpp\"");

		return importClasses;
	}


	// Generate a set of standard classes for import statements
	private List<String> getAllLibClasses(Collection<String> stdLibClasses, Collection<String> libClasses) {

		List<String> allLibClasses = new ArrayList<String>(stdLibClasses);
		// Iterate over the list of import classes
		for (String str : libClasses) {
			if (!allLibClasses.contains(str)) {
				allLibClasses.add(str);
			}
		}

		return allLibClasses;
	}



	// Generate a set of classes for import statements
	private Set<String> getImportClasses(Collection<String> methods, InterfaceDecl intDecl) {

		Set<String> importClasses = new HashSet<String>();
		for (String method : methods) {
			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			for (String paramType : methPrmTypes) {

				String simpleType = getSimpleType(paramType);
				if (getParamCategory(simpleType) == ParamCategory.NONPRIMITIVES) {
					importClasses.add(getNonPrimitiveJavaClass(simpleType));
				}
			}
		}
		return importClasses;
	}


	// Generate a set of classes for include statements
	private Set<String> getIncludeClasses(Collection<String> methods, InterfaceDecl intDecl, boolean needExchange) {

		Set<String> includeClasses = new HashSet<String>();
		for (String method : methods) {

			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			List<String> methParams = intDecl.getMethodParams(method);
			for (int i = 0; i < methPrmTypes.size(); i++) {

				String simpleType = getSimpleType(methPrmTypes.get(i));
				String param = methParams.get(i);
				if (getParamCategory(simpleType) == ParamCategory.NONPRIMITIVES) {
					includeClasses.add("<" + getNonPrimitiveCplusClass(simpleType) + ">");
				} else if (getParamCategory(simpleType) == ParamCategory.USERDEFINED) {
					// For original interface, we need it exchanged... not for stub interfaces
					if (needExchange) {
						includeClasses.add("\"" + exchangeParamType(simpleType) + ".hpp\"");
						includeClasses.add("\"" + exchangeParamType(simpleType) + "_CallbackStub.hpp\"");
					} else {
						includeClasses.add("\"" + simpleType + ".hpp\"");
						includeClasses.add("\"" + simpleType + "_CallbackSkeleton.hpp\"");
					}
				} else if (param.contains("[]")) {
				// Check if this is array for C++; translate into vector
					includeClasses.add("<vector>");
				}
			}
		}
		return includeClasses;
	}


	// Generate a set of callback classes
	private Set<String> getCallbackClasses(Collection<String> methods, InterfaceDecl intDecl) {

		Set<String> callbackClasses = new HashSet<String>();
		for (String method : methods) {

			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			List<String> methParams = intDecl.getMethodParams(method);
			for (int i = 0; i < methPrmTypes.size(); i++) {

				String type = methPrmTypes.get(i);
				if (getParamCategory(type) == ParamCategory.USERDEFINED) {
					callbackClasses.add(type);
				} else if (getParamCategory(type) == ParamCategory.NONPRIMITIVES) {
				// Can be a List<...> of callback objects ...
					String genericType = getTypeOfGeneric(type)[0];
					if (getParamCategory(type) == ParamCategory.USERDEFINED) {
						callbackClasses.add(type);
					}
				}
			}
		}
		return callbackClasses;
	}


	private void printImportStatements(Collection<String> importClasses) {

		for(String cls : importClasses) {
			println("import " + cls + ";");
		}
	}


	private void printIncludeStatements(Collection<String> includeClasses) {

		for(String cls : includeClasses) {
			println("#include " + cls);
		}
	}


	// Get the C++ version of a non-primitive type
	// e.g. set for Set and map for Map
	// Input nonPrimitiveType has to be generics in format
	private String[] getTypeOfGeneric(String nonPrimitiveType) {

		// Handle <, >, and , for 2-type generic/template
		String[] substr = nonPrimitiveType.split("<")[1].split(">")[0].split(",");
		return substr;
	}


	// This helper function strips off array declaration, e.g. D[] becomes D
	private String getSimpleIdentifier(String ident) {

		// Handle [ for array declaration
		String substr = ident;
		if (ident.contains("[]")) {
			substr = ident.split("\\[\\]")[0];
		}
		return substr;
	}


	private String checkAndGetCplusType(String paramType) {

		if (getParamCategory(paramType) == ParamCategory.PRIMITIVES) {
			return convertType(paramType);
		} else if (getParamCategory(paramType) == ParamCategory.NONPRIMITIVES) {

			// Check for generic/template format
			if (paramType.contains("<") && paramType.contains(">")) {

				String genericClass = getSimpleType(paramType);
				String[] genericType = getTypeOfGeneric(paramType);
				String cplusTemplate = null;
				if (genericType.length == 1) // Generic/template with one type
					cplusTemplate = getNonPrimitiveCplusClass(genericClass) + 
						"<" + convertType(genericType[0]) + ">";
				else // Generic/template with two types
					cplusTemplate = getNonPrimitiveCplusClass(genericClass) + 
						"<" + convertType(genericType[0]) + "," + convertType(genericType[1]) + ">";
				return cplusTemplate;
			} else
				return getNonPrimitiveCplusClass(paramType);
		} else if(getParamCategory(paramType) == ParamCategory.USERDEFINED) {
			return paramType + "*";
		} else
			// Just return it as is if it's not non-primitives
			return paramType;
			//return checkAndGetParamClass(paramType, true);
	}


	// Detect array declaration, e.g. int A[],
	// 		then generate "int A[]" in C++ as "vector<int> A"
	private String checkAndGetCplusArray(String paramType, String param) {

		String paramComplete = null;
		// Check for array declaration
		if (param.contains("[]")) {
			paramComplete = "vector<" + paramType + "> " + param.replace("[]","");
		} else
			// Just return it as is if it's not an array
			paramComplete = paramType + " " + param;

		return paramComplete;
	}
	

	// Detect array declaration, e.g. int A[],
	// 		then generate "int A[]" in C++ as "vector<int> A"
	// This method just returns the type
	private String checkAndGetCplusArrayType(String paramType) {

		String paramTypeRet = null;
		// Check for array declaration
		if (paramType.contains("[]")) {
			String type = paramType.split("\\[\\]")[0];
			paramTypeRet = checkAndGetCplusType(type) + "[]";
		} else if (paramType.contains("vector")) {
			// Just return it as is if it's not an array
			String type = paramType.split("<")[1].split(">")[0];
			paramTypeRet = checkAndGetCplusType(type) + "[]";
		} else
			paramTypeRet = paramType;

		return paramTypeRet;
	}
	
	
	// Detect array declaration, e.g. int A[],
	// 		then generate "int A[]" in C++ as "vector<int> A"
	// This method just returns the type
	private String checkAndGetCplusArrayType(String paramType, String param) {

		String paramTypeRet = null;
		// Check for array declaration
		if (param.contains("[]")) {
			paramTypeRet = checkAndGetCplusType(paramType) + "[]";
		} else if (paramType.contains("vector")) {
			// Just return it as is if it's not an array
			String type = paramType.split("<")[1].split(">")[0];
			paramTypeRet = checkAndGetCplusType(type) + "[]";
		} else
			paramTypeRet = paramType;

		return paramTypeRet;
	}


	// Detect array declaration, e.g. int A[],
	// 		then generate type "int[]"
	private String checkAndGetArray(String paramType, String param) {

		String paramTypeRet = null;
		// Check for array declaration
		if (param.contains("[]")) {
			paramTypeRet = paramType + "[]";
		} else
			// Just return it as is if it's not an array
			paramTypeRet = paramType;

		return paramTypeRet;
	}


	// Is array or list?
	private boolean isArrayOrList(String paramType, String param) {

		// Check for array declaration
		if (isArray(paramType, param))
			return true;
		else if (isList(paramType, param))
			return true;
		else
			return false;
	}


	// Is array or list?
	private boolean isArray(String paramType, String param) {

		// Check for array declaration
		if (param.contains("[]"))
			return true;
		else
			return false;
	}


	// Is array or list?
	private boolean isList(String paramType, String param) {

		// Check for array declaration
		if (paramType.contains("List"))
			return true;
		else
			return false;
	}


	// Get the right type for a callback object
	private String checkAndGetParamClass(String paramType) {

		// Check if this is generics
		if(getParamCategory(paramType) == ParamCategory.USERDEFINED) {
			return exchangeParamType(paramType);
		} else
			return paramType;
	}


	// Returns the other interface for type-checking purposes for USERDEFINED
	//		classes based on the information provided in multiple policy files
	// e.g. return CameraWithXXX instead of Camera
	private String exchangeParamType(String intface) {

		// Param type that's passed is the interface name we need to look for
		//		in the map of interfaces, based on available policy files.
		DeclarationHandler decHandler = mapIntDeclHand.get(intface);
		if (decHandler != null) {
		// We've found the required interface policy files
			RequiresDecl reqDecl = (RequiresDecl) decHandler.getRequiresDecl(intface);
			Set<String> setExchInt = reqDecl.getInterfaces();
			if (setExchInt.size() == 1) {
				Iterator iter = setExchInt.iterator();
				return (String) iter.next();
			} else {
				throw new Error("IoTCompiler: Ambiguous stub interfaces: " + setExchInt.toString() + 
					". Only one new interface can be declared if the object " + intface +
					" needs to be passed in as an input parameter!");
			}
		} else {
		// NULL value - this means policy files missing
			throw new Error("IoTCompiler: Parameter type lookup failed for " + intface +
				"... Please provide the necessary policy files for user-defined types." +
				" If this is an array please type the brackets after the variable name," +
				" e.g. \"String str[]\", not \"String[] str\"." +
				" If this is a Collections (Java) / STL (C++) type, this compiler only" +
				" supports List/ArrayList (Java) or list (C++).");
		}
	}


	public static void main(String[] args) throws Exception {

		// If there is no argument or just "--help" or "-h", then invoke printUsage()
		if ((args[0].equals("-help") ||
			 args[0].equals("--help")||
			 args[0].equals("-h"))   ||
			(args.length == 0)) {

			IoTCompiler.printUsage();

		} else if (args.length > 1) {

			IoTCompiler comp = new IoTCompiler();
			int i = 0;				
			do {
				// Parse main policy file
				ParseNode pnPol = IoTCompiler.parseFile(args[i]);
				// Parse "requires" policy file
				ParseNode pnReq = IoTCompiler.parseFile(args[i+1]);
				// Get interface name
				String intface = ParseTreeHandler.getOrigIntface(pnPol);
				comp.setDataStructures(intface, pnPol, pnReq);
				comp.getMethodsForIntface(intface);
				i = i + 2;
			// 1) Check if this is the last option before "-java" or "-cplus"
			// 2) Check if this is really the last option (no "-java" or "-cplus")
			} while(!args[i].equals("-java") &&
					!args[i].equals("-cplus") &&
					(i < args.length));

			// Generate everything if we don't see "-java" or "-cplus"
			if (i == args.length) {
				comp.generateJavaLocalInterfaces();
				comp.generateJavaInterfaces();
				comp.generateJavaStubClasses();
				comp.generateJavaCallbackStubClasses();
				comp.generateJavaSkeletonClass();
				comp.generateJavaCallbackSkeletonClass();
				comp.generateCplusLocalInterfaces();
				comp.generateCPlusInterfaces();
				comp.generateCPlusStubClasses();
				comp.generateCPlusCallbackStubClasses();
				comp.generateCplusSkeletonClass();
				comp.generateCplusCallbackSkeletonClass();
			} else {
			// Check other options
				while(i < args.length) {
					// Error checking
					if (!args[i].equals("-java") &&
						!args[i].equals("-cplus")) {
						throw new Error("IoTCompiler: ERROR - unrecognized command line option: " + args[i]);
					} else {
						if (i + 1 < args.length) {
							comp.setDirectory(args[i+1]);
						} else
							throw new Error("IoTCompiler: ERROR - please provide <directory> after option: " + args[i]);

						if (args[i].equals("-java")) {
							comp.generateJavaLocalInterfaces();
							comp.generateJavaInterfaces();
							comp.generateJavaStubClasses();
							comp.generateJavaCallbackStubClasses();
							comp.generateJavaSkeletonClass();
							comp.generateJavaCallbackSkeletonClass();
						} else {
							comp.generateCplusLocalInterfaces();
							comp.generateCPlusInterfaces();
							comp.generateCPlusStubClasses();
							comp.generateCPlusCallbackStubClasses();
							comp.generateCplusSkeletonClass();
							comp.generateCplusCallbackSkeletonClass();
						}
					}
					i = i + 2;
				}
			}
		} else {
		// Need to at least have exactly 2 parameters, i.e. main policy file and requires file
			IoTCompiler.printUsage();
			throw new Error("IoTCompiler: At least two arguments (main and requires policy files) have to be provided!");
		}
	}
}




